"""
trionyx.trionyx.forms
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

:copyright: 2018 by Maikel Martens
:license: GPLv3
"""
from django.utils.translation import ugettext_lazy as _
from django.contrib.auth import password_validation
from django.contrib.auth.models import Group
from django.contrib.contenttypes.models import ContentType

from trionyx.models import get_class
from trionyx import forms
from trionyx.forms.layout import Layout, Fieldset, Div, HtmlTemplate, Filters
from trionyx.forms.helper import FormHelper
from trionyx.trionyx.icons import ICON_CHOICES
from trionyx.config import models_config
from trionyx import utils

User = get_class('trionyx.User')


class ProfileUpdateForm(forms.ModelForm):
    """User update form"""

    error_messages = {
        'password_mismatch': _("The two password fields didn't match."),
    }
    new_password1 = forms.CharField(
        label=_("New password"),
        widget=forms.PasswordInput,
        strip=False,
        help_text=password_validation.password_validators_help_text_html(),
        required=False,
    )
    new_password2 = forms.CharField(
        label=_("New password confirmation"),
        strip=False,
        widget=forms.PasswordInput,
        required=False,
    )

    def __init__(self, instance=None, *args, **kwargs):
        """Init user form"""
        super().__init__(*args, instance=instance, **kwargs)
        self.fields['language'].choices = [(key, _(value)) for key, value in self.fields['language'].choices]
        self.helper = FormHelper()
        self.helper.layout = Layout(
            'email',
            Div(
                Fieldset(
                    _('Personal info'),
                    'first_name',
                    'last_name',
                    css_class="col-md-6",
                ),
                Div(
                    'avatar',
                    css_class="col-md-6",
                ),
                css_class="row"
            ),
            Div(
                Fieldset(
                    _('Change password'),
                    'new_password1',
                    'new_password2',
                    css_class='col-md-6'
                ),
                Fieldset(
                    _('Settings'),
                    'language',
                    'timezone',
                    css_class='col-md-6'
                ),
                css_class='row'
            ),
        )

    class Meta:
        """Meta description for form"""

        model = User
        fields = ['new_password1', 'new_password2', 'email', 'first_name', 'last_name', 'avatar', 'language', 'timezone']

    def clean_new_password2(self):
        """Validate password when set"""
        password1 = self.cleaned_data.get('new_password1')
        password2 = self.cleaned_data.get('new_password2')
        if password1 or password2:
            if password1 != password2:
                raise forms.ValidationError(
                    self.error_messages['password_mismatch'],
                    code='password_mismatch',
                )
            password_validation.validate_password(password2, self.instance)
        return password2

    def save(self, commit=True):
        """Save user"""
        user = super().save(commit=False)

        password = self.cleaned_data["new_password1"]
        if password:
            user.set_password(password)

        if commit:
            user.save()
        return user


@forms.register(default_create=True)
class UserCreateForm(forms.ModelForm):
    """User update form"""

    error_messages = {
        'password_mismatch': _("The two password fields didn't match."),
    }
    new_password1 = forms.CharField(
        label=_("New password"),
        widget=forms.PasswordInput,
        strip=False,
        help_text=password_validation.password_validators_help_text_html(),
        required=True,
    )
    new_password2 = forms.CharField(
        label=_("New password confirmation"),
        strip=False,
        widget=forms.PasswordInput,
        required=True,
    )

    def __init__(self, instance=None, *args, **kwargs):
        """Init user form"""
        super().__init__(*args, instance=instance, **kwargs)
        self.helper = FormHelper()
        self.helper.layout = Layout(
            Div(
                Fieldset(
                    _('Login'),
                    'email',
                    'new_password1',
                    'new_password2',
                    css_class="col-md-6",
                ),
                Fieldset(
                    _('Personal info'),
                    'first_name',
                    'last_name',
                    css_class="col-md-6",
                ),
                css_class="row"
            ),

        )

    class Meta:
        """Meta description for form"""

        model = User
        fields = ['new_password1', 'new_password2', 'email', 'first_name', 'last_name']

    def clean_new_password2(self):
        """Validate password when set"""
        password1 = self.cleaned_data.get('new_password1')
        password2 = self.cleaned_data.get('new_password2')
        if password1 != password2:
            raise forms.ValidationError(
                self.error_messages['password_mismatch'],
                code='password_mismatch',
            )
        password_validation.validate_password(password2, self.instance)
        return password2

    def save(self, commit=True):
        """Save user"""
        user = super().save(commit=False)

        password = self.cleaned_data["new_password1"]
        if password:
            user.set_password(password)

        if commit:
            user.save()
        return user


@forms.register(default_edit=True)
class UserUpdateForm(forms.ModelForm):
    """User update form"""

    css_files = [
        'plugins/jstree/themes/default/style.css'
    ]

    js_files = [
        'plugins/jstree/jstree.min.js'
    ]

    error_messages = {
        'password_mismatch': _("The two password fields didn't match."),
    }
    new_password1 = forms.CharField(
        label=_("New password"),
        widget=forms.PasswordInput,
        strip=False,
        help_text=password_validation.password_validators_help_text_html(),
        required=False,
    )
    new_password2 = forms.CharField(
        label=_("New password confirmation"),
        strip=False,
        widget=forms.PasswordInput,
        required=False,
    )

    def __init__(self, instance=None, *args, **kwargs):
        """Init user form"""
        super().__init__(*args, instance=instance, **kwargs)
        from trionyx.trionyx.views import create_permission_jstree

        selected_permissions = []
        if self.instance:
            selected_permissions = self.instance.user_permissions.all()
        elif 'user_permissions' in self.initial:
            selected_permissions = self.initial['user_permissions']

        jstree = create_permission_jstree(selected_permissions)

        self.helper = FormHelper()
        self.helper.layout = Layout(
            'email',
            Div(
                Div(
                    Fieldset(
                        _('Personal info'),
                        'first_name',
                        'last_name',
                        'avatar',
                    ),
                    Fieldset(
                        _('Change password'),
                        'new_password1',
                        'new_password2',
                    ),
                    css_class="col-md-6",
                ),
                Fieldset(
                    _('Permissions'),
                    'is_active',
                    'is_superuser',
                    'groups',
                    HtmlTemplate('trionyx/forms/permissions.html', {
                        'field_name': 'user_permissions',
                        'permission_jstree': jstree,
                        'permission_ids': filter(None, [
                            item['permission_id'] for item in jstree if 'permission_id' in item]),
                    }),
                    css_class="col-md-6",
                ),
                css_class="row"
            ),
        )

    class Meta:
        """Meta description for form"""

        model = User
        fields = [
            'new_password1', 'new_password2', 'email', 'first_name', 'last_name',
            'avatar', 'user_permissions', 'groups', 'is_superuser', 'is_active'
        ]

    def clean_new_password2(self):
        """Validate password when set"""
        password1 = self.cleaned_data.get('new_password1')
        password2 = self.cleaned_data.get('new_password2')
        if password1 or password2:
            if password1 != password2:
                raise forms.ValidationError(
                    self.error_messages['password_mismatch'],
                    code='password_mismatch',
                )
            password_validation.validate_password(password2, self.instance)
        return password2

    def save(self, commit=True):
        """Save user"""
        user = super().save(commit=False)

        password = self.cleaned_data["new_password1"]
        if password:
            user.set_password(password)

        if commit:
            user.save()
            self.save_m2m()
        return user


@forms.register(default_create=True, default_edit=True)
class GroupForm(forms.ModelForm):
    """Group form"""

    css_files = [
        'plugins/jstree/themes/default/style.css'
    ]

    js_files = [
        'plugins/jstree/jstree.min.js'
    ]

    def __init__(self, *args, **kwargs):
        """Init GroupForm"""
        super().__init__(*args, **kwargs)  # populates the post
        from trionyx.trionyx.views import create_permission_jstree

        selected_permissions = []
        if self.instance:
            try:
                selected_permissions = self.instance.permissions.all()
            except ValueError:
                if 'permissions' in self.initial:
                    selected_permissions = self.initial['permissions']
        elif 'permissions' in self.initial:
            selected_permissions = self.initial['permissions']

        jstree = create_permission_jstree(selected_permissions)

        self.helper = FormHelper()
        self.helper.layout = Layout(
            'name',
            HtmlTemplate('trionyx/forms/permissions.html', {
                'field_name': 'permissions',
                'permission_jstree': jstree,
                'permission_ids': filter(None, [item['permission_id'] for item in jstree if 'permission_id' in item]),
            })
        )

    class Meta:
        """Meta description for form"""

        model = Group
        fields = ['name', 'permissions']


class AuditlogWidgetForm(forms.Form):
    """Auditlog widget form"""

    SHOW_CHOICES = [
        ('all', _('All')),
        ('user', _('Users only')),
        ('system', _('Systems only')),
    ]

    show = forms.ChoiceField(label=_('Show'), choices=SHOW_CHOICES)


class TotalSummaryWidgetForm(forms.Form):
    """Total summary widget form"""

    PERIOD_CHOICES = [
        ('all', _('All')),
        ('year', _('Current year')),
        ('month', _('Current month')),
        ('week', _('Current week')),
        ('day', _('Current day')),
        ('365days', _('Last 365 days')),
        ('30days', _('Last 30 days')),
        ('7days', _('Last 7 days')),
    ]

    icon = forms.ChoiceField(label=_('Icon'), choices=ICON_CHOICES)

    model = forms.ChoiceField(label=_('Model'), choices=[], required=False)
    field = forms.CharField(label=_('Field'))

    period = forms.ChoiceField(label=_('Period'), choices=PERIOD_CHOICES)
    period_field = forms.CharField(label=_('Period field'), required=False)
    filters = forms.CharField(label=_('Filters'), required=False)

    def __init__(self, *args, **kwargs):
        """Init form"""
        super().__init__(*args, **kwargs)

        # TODO Change this to use ajax
        content_type_map = ContentType.objects.get_for_models(*list(models_config.get_all_models(utils.get_current_request().user)))
        period_fields = {}
        summary_fields = {}

        for config in models_config.get_all_configs():
            if config.model not in content_type_map:
                continue

            summary_fields[content_type_map[config.model].id] = [
                {
                    'id': field.name,
                    'text': str(field.verbose_name)
                } for field in config.get_fields() if config.get_field_type(field) in ['int', 'float']
            ]

            period_fields[content_type_map[config.model].id] = [
                {
                    'id': field.name,
                    'text': str(field.verbose_name)
                } for field in config.get_fields(True) if config.get_field_type(field) in ['date', 'datetime']
            ]

        self.fields['model'].choices = [
            (content_type.id, model._meta.verbose_name.capitalize()) for model, content_type
            in content_type_map.items()]

        self.helper = FormHelper()
        self.helper.layout = Layout(
            'icon',
            Div(
                Div(
                    'model',
                    css_class="col-md-6",
                ),
                Div(
                    HtmlTemplate('widgets/total_summary/form_field.html', {
                        'summary_fields': summary_fields,
                        'value': self.initial.get('field'),
                    }),
                    css_class="col-md-6",
                ),
                css_class="row"
            ),
            Div(
                Div(
                    'period',
                    css_class="col-md-6",
                ),
                Div(
                    HtmlTemplate('widgets/total_summary/period_field.html', {
                        'period_fields': period_fields,
                        'value': self.initial.get('period_field'),
                    }),
                    css_class="col-md-6",
                ),
                css_class="row"
            ),
            Fieldset(
                _('Filters'),
                Filters('filters', content_type_input_id='id_model'),
            )
        )



"""
trionyx.trionyx.icons
~~~~~~~~~~~~~~~~~~~~~

:copyright: 2019 by Maikel Martens
:license: GPLv3
"""
ICON_CHOICES = [
    ('fa fa-dashboard', '<i class="fa fa-dashboard"></i> Dashboard'),
    ('fa fa-500px', '<i class="fa fa-500px"></i> 500px'),
    ('fa fa-amazon', '<i class="fa fa-amazon"></i> Amazon'),
    ('fa fa-balance-scale', '<i class="fa fa-balance-scale"></i> Balance-scale'),
    ('fa fa-battery-0', '<i class="fa fa-battery-0"></i> Battery-0'),
    ('fa fa-battery-1', '<i class="fa fa-battery-1"></i> Battery-1'),
    ('fa fa-battery-2', '<i class="fa fa-battery-2"></i> Battery-2'),
    ('fa fa-battery-3', '<i class="fa fa-battery-3"></i> Battery-3'),
    ('fa fa-battery-4', '<i class="fa fa-battery-4"></i> Battery-4'),
    ('fa fa-battery-empty', '<i class="fa fa-battery-empty"></i> Battery-empty'),
    ('fa fa-battery-full', '<i class="fa fa-battery-full"></i> Battery-full'),
    ('fa fa-battery-half', '<i class="fa fa-battery-half"></i> Battery-half'),
    ('fa fa-battery-quarter', '<i class="fa fa-battery-quarter"></i> Battery-quarter'),
    ('fa fa-battery-three-quarters', '<i class="fa fa-battery-three-quarters"></i> Battery-three-quarters'),
    ('fa fa-black-tie', '<i class="fa fa-black-tie"></i> Black-tie'),
    ('fa fa-calendar-check-o', '<i class="fa fa-calendar-check-o"></i> Calendar-check-o'),
    ('fa fa-calendar-minus-o', '<i class="fa fa-calendar-minus-o"></i> Calendar-minus-o'),
    ('fa fa-calendar-plus-o', '<i class="fa fa-calendar-plus-o"></i> Calendar-plus-o'),
    ('fa fa-calendar-times-o', '<i class="fa fa-calendar-times-o"></i> Calendar-times-o'),
    ('fa fa-cc-diners-club', '<i class="fa fa-cc-diners-club"></i> Cc-diners-club'),
    ('fa fa-cc-jcb', '<i class="fa fa-cc-jcb"></i> Cc-jcb'),
    ('fa fa-chrome', '<i class="fa fa-chrome"></i> Chrome'),
    ('fa fa-clone', '<i class="fa fa-clone"></i> Clone'),
    ('fa fa-commenting', '<i class="fa fa-commenting"></i> Commenting'),
    ('fa fa-commenting-o', '<i class="fa fa-commenting-o"></i> Commenting-o'),
    ('fa fa-contao', '<i class="fa fa-contao"></i> Contao'),
    ('fa fa-creative-commons', '<i class="fa fa-creative-commons"></i> Creative-commons'),
    ('fa fa-expeditedssl', '<i class="fa fa-expeditedssl"></i> Expeditedssl'),
    ('fa fa-firefox', '<i class="fa fa-firefox"></i> Firefox'),
    ('fa fa-fonticons', '<i class="fa fa-fonticons"></i> Fonticons'),
    ('fa fa-genderless', '<i class="fa fa-genderless"></i> Genderless'),
    ('fa fa-get-pocket', '<i class="fa fa-get-pocket"></i> Get-pocket'),
    ('fa fa-gg', '<i class="fa fa-gg"></i> Gg'),
    ('fa fa-gg-circle', '<i class="fa fa-gg-circle"></i> Gg-circle'),
    ('fa fa-hand-grab-o', '<i class="fa fa-hand-grab-o"></i> Hand-grab-o'),
    ('fa fa-hand-lizard-o', '<i class="fa fa-hand-lizard-o"></i> Hand-lizard-o'),
    ('fa fa-hand-paper-o', '<i class="fa fa-hand-paper-o"></i> Hand-paper-o'),
    ('fa fa-hand-peace-o', '<i class="fa fa-hand-peace-o"></i> Hand-peace-o'),
    ('fa fa-hand-pointer-o', '<i class="fa fa-hand-pointer-o"></i> Hand-pointer-o'),
    ('fa fa-hand-rock-o', '<i class="fa fa-hand-rock-o"></i> Hand-rock-o'),
    ('fa fa-hand-scissors-o', '<i class="fa fa-hand-scissors-o"></i> Hand-scissors-o'),
    ('fa fa-hand-spock-o', '<i class="fa fa-hand-spock-o"></i> Hand-spock-o'),
    ('fa fa-hand-stop-o', '<i class="fa fa-hand-stop-o"></i> Hand-stop-o'),
    ('fa fa-hourglass', '<i class="fa fa-hourglass"></i> Hourglass'),
    ('fa fa-hourglass-1', '<i class="fa fa-hourglass-1"></i> Hourglass-1'),
    ('fa fa-hourglass-2', '<i class="fa fa-hourglass-2"></i> Hourglass-2'),
    ('fa fa-hourglass-3', '<i class="fa fa-hourglass-3"></i> Hourglass-3'),
    ('fa fa-hourglass-end', '<i class="fa fa-hourglass-end"></i> Hourglass-end'),
    ('fa fa-hourglass-half', '<i class="fa fa-hourglass-half"></i> Hourglass-half'),
    ('fa fa-hourglass-o', '<i class="fa fa-hourglass-o"></i> Hourglass-o'),
    ('fa fa-hourglass-start', '<i class="fa fa-hourglass-start"></i> Hourglass-start'),
    ('fa fa-houzz', '<i class="fa fa-houzz"></i> Houzz'),
    ('fa fa-i-cursor', '<i class="fa fa-i-cursor"></i> I-cursor'),
    ('fa fa-industry', '<i class="fa fa-industry"></i> Industry'),
    ('fa fa-internet-explorer', '<i class="fa fa-internet-explorer"></i> Internet-explorer'),
    ('fa fa-map', '<i class="fa fa-map"></i> Map'),
    ('fa fa-map-o', '<i class="fa fa-map-o"></i> Map-o'),
    ('fa fa-map-pin', '<i class="fa fa-map-pin"></i> Map-pin'),
    ('fa fa-map-signs', '<i class="fa fa-map-signs"></i> Map-signs'),
    ('fa fa-mouse-pointer', '<i class="fa fa-mouse-pointer"></i> Mouse-pointer'),
    ('fa fa-object-group', '<i class="fa fa-object-group"></i> Object-group'),
    ('fa fa-object-ungroup', '<i class="fa fa-object-ungroup"></i> Object-ungroup'),
    ('fa fa-odnoklassniki', '<i class="fa fa-odnoklassniki"></i> Odnoklassniki'),
    ('fa fa-odnoklassniki-square', '<i class="fa fa-odnoklassniki-square"></i> Odnoklassniki-square'),
    ('fa fa-opencart', '<i class="fa fa-opencart"></i> Opencart'),
    ('fa fa-opera', '<i class="fa fa-opera"></i> Opera'),
    ('fa fa-optin-monster', '<i class="fa fa-optin-monster"></i> Optin-monster'),
    ('fa fa-registered', '<i class="fa fa-registered"></i> Registered'),
    ('fa fa-safari', '<i class="fa fa-safari"></i> Safari'),
    ('fa fa-sticky-note', '<i class="fa fa-sticky-note"></i> Sticky-note'),
    ('fa fa-sticky-note-o', '<i class="fa fa-sticky-note-o"></i> Sticky-note-o'),
    ('fa fa-television', '<i class="fa fa-television"></i> Television'),
    ('fa fa-trademark', '<i class="fa fa-trademark"></i> Trademark'),
    ('fa fa-tripadvisor', '<i class="fa fa-tripadvisor"></i> Tripadvisor'),
    ('fa fa-tv', '<i class="fa fa-tv"></i> Tv'),
    ('fa fa-vimeo', '<i class="fa fa-vimeo"></i> Vimeo'),
    ('fa fa-wikipedia-w', '<i class="fa fa-wikipedia-w"></i> Wikipedia-w'),
    ('fa fa-y-combinator', '<i class="fa fa-y-combinator"></i> Y-combinator'),
    ('fa fa-yc', '<i class="fa fa-yc"></i> Yc'),
    ('fa fa-adjust', '<i class="fa fa-adjust"></i> Adjust'),
    ('fa fa-anchor', '<i class="fa fa-anchor"></i> Anchor'),
    ('fa fa-archive', '<i class="fa fa-archive"></i> Archive'),
    ('fa fa-area-chart', '<i class="fa fa-area-chart"></i> Area-chart'),
    ('fa fa-arrows', '<i class="fa fa-arrows"></i> Arrows'),
    ('fa fa-arrows-h', '<i class="fa fa-arrows-h"></i> Arrows-h'),
    ('fa fa-arrows-v', '<i class="fa fa-arrows-v"></i> Arrows-v'),
    ('fa fa-asterisk', '<i class="fa fa-asterisk"></i> Asterisk'),
    ('fa fa-at', '<i class="fa fa-at"></i> At'),
    ('fa fa-automobile', '<i class="fa fa-automobile"></i> Automobile'),
    ('fa fa-ban', '<i class="fa fa-ban"></i> Ban'),
    ('fa fa-bank', '<i class="fa fa-bank"></i> Bank'),
    ('fa fa-bar-chart', '<i class="fa fa-bar-chart"></i> Bar-chart'),
    ('fa fa-bar-chart-o', '<i class="fa fa-bar-chart-o"></i> Bar-chart-o'),
    ('fa fa-barcode', '<i class="fa fa-barcode"></i> Barcode'),
    ('fa fa-bars', '<i class="fa fa-bars"></i> Bars'),
    ('fa fa-bed', '<i class="fa fa-bed"></i> Bed'),
    ('fa fa-beer', '<i class="fa fa-beer"></i> Beer'),
    ('fa fa-bell', '<i class="fa fa-bell"></i> Bell'),
    ('fa fa-bell-o', '<i class="fa fa-bell-o"></i> Bell-o'),
    ('fa fa-bell-slash', '<i class="fa fa-bell-slash"></i> Bell-slash'),
    ('fa fa-bell-slash-o', '<i class="fa fa-bell-slash-o"></i> Bell-slash-o'),
    ('fa fa-bicycle', '<i class="fa fa-bicycle"></i> Bicycle'),
    ('fa fa-binoculars', '<i class="fa fa-binoculars"></i> Binoculars'),
    ('fa fa-birthday-cake', '<i class="fa fa-birthday-cake"></i> Birthday-cake'),
    ('fa fa-bolt', '<i class="fa fa-bolt"></i> Bolt'),
    ('fa fa-bomb', '<i class="fa fa-bomb"></i> Bomb'),
    ('fa fa-book', '<i class="fa fa-book"></i> Book'),
    ('fa fa-bookmark', '<i class="fa fa-bookmark"></i> Bookmark'),
    ('fa fa-bookmark-o', '<i class="fa fa-bookmark-o"></i> Bookmark-o'),
    ('fa fa-briefcase', '<i class="fa fa-briefcase"></i> Briefcase'),
    ('fa fa-bug', '<i class="fa fa-bug"></i> Bug'),
    ('fa fa-building', '<i class="fa fa-building"></i> Building'),
    ('fa fa-building-o', '<i class="fa fa-building-o"></i> Building-o'),
    ('fa fa-bullhorn', '<i class="fa fa-bullhorn"></i> Bullhorn'),
    ('fa fa-bullseye', '<i class="fa fa-bullseye"></i> Bullseye'),
    ('fa fa-bus', '<i class="fa fa-bus"></i> Bus'),
    ('fa fa-cab', '<i class="fa fa-cab"></i> Cab'),
    ('fa fa-calculator', '<i class="fa fa-calculator"></i> Calculator'),
    ('fa fa-calendar', '<i class="fa fa-calendar"></i> Calendar'),
    ('fa fa-calendar-o', '<i class="fa fa-calendar-o"></i> Calendar-o'),
    ('fa fa-camera', '<i class="fa fa-camera"></i> Camera'),
    ('fa fa-camera-retro', '<i class="fa fa-camera-retro"></i> Camera-retro'),
    ('fa fa-car', '<i class="fa fa-car"></i> Car'),
    ('fa fa-caret-square-o-down', '<i class="fa fa-caret-square-o-down"></i> Caret-square-o-down'),
    ('fa fa-caret-square-o-left', '<i class="fa fa-caret-square-o-left"></i> Caret-square-o-left'),
    ('fa fa-caret-square-o-right', '<i class="fa fa-caret-square-o-right"></i> Caret-square-o-right'),
    ('fa fa-caret-square-o-up', '<i class="fa fa-caret-square-o-up"></i> Caret-square-o-up'),
    ('fa fa-cart-arrow-down', '<i class="fa fa-cart-arrow-down"></i> Cart-arrow-down'),
    ('fa fa-cart-plus', '<i class="fa fa-cart-plus"></i> Cart-plus'),
    ('fa fa-cc', '<i class="fa fa-cc"></i> Cc'),
    ('fa fa-certificate', '<i class="fa fa-certificate"></i> Certificate'),
    ('fa fa-check', '<i class="fa fa-check"></i> Check'),
    ('fa fa-check-circle', '<i class="fa fa-check-circle"></i> Check-circle'),
    ('fa fa-check-circle-o', '<i class="fa fa-check-circle-o"></i> Check-circle-o'),
    ('fa fa-check-square', '<i class="fa fa-check-square"></i> Check-square'),
    ('fa fa-check-square-o', '<i class="fa fa-check-square-o"></i> Check-square-o'),
    ('fa fa-child', '<i class="fa fa-child"></i> Child'),
    ('fa fa-circle', '<i class="fa fa-circle"></i> Circle'),
    ('fa fa-circle-o', '<i class="fa fa-circle-o"></i> Circle-o'),
    ('fa fa-circle-o-notch', '<i class="fa fa-circle-o-notch"></i> Circle-o-notch'),
    ('fa fa-circle-thin', '<i class="fa fa-circle-thin"></i> Circle-thin'),
    ('fa fa-clock-o', '<i class="fa fa-clock-o"></i> Clock-o'),
    ('fa fa-close', '<i class="fa fa-close"></i> Close'),
    ('fa fa-cloud', '<i class="fa fa-cloud"></i> Cloud'),
    ('fa fa-cloud-download', '<i class="fa fa-cloud-download"></i> Cloud-download'),
    ('fa fa-cloud-upload', '<i class="fa fa-cloud-upload"></i> Cloud-upload'),
    ('fa fa-code', '<i class="fa fa-code"></i> Code'),
    ('fa fa-code-fork', '<i class="fa fa-code-fork"></i> Code-fork'),
    ('fa fa-coffee', '<i class="fa fa-coffee"></i> Coffee'),
    ('fa fa-cog', '<i class="fa fa-cog"></i> Cog'),
    ('fa fa-cogs', '<i class="fa fa-cogs"></i> Cogs'),
    ('fa fa-comment', '<i class="fa fa-comment"></i> Comment'),
    ('fa fa-comment-o', '<i class="fa fa-comment-o"></i> Comment-o'),
    ('fa fa-comments', '<i class="fa fa-comments"></i> Comments'),
    ('fa fa-comments-o', '<i class="fa fa-comments-o"></i> Comments-o'),
    ('fa fa-compass', '<i class="fa fa-compass"></i> Compass'),
    ('fa fa-copyright', '<i class="fa fa-copyright"></i> Copyright'),
    ('fa fa-credit-card', '<i class="fa fa-credit-card"></i> Credit-card'),
    ('fa fa-crop', '<i class="fa fa-crop"></i> Crop'),
    ('fa fa-crosshairs', '<i class="fa fa-crosshairs"></i> Crosshairs'),
    ('fa fa-cube', '<i class="fa fa-cube"></i> Cube'),
    ('fa fa-cubes', '<i class="fa fa-cubes"></i> Cubes'),
    ('fa fa-cutlery', '<i class="fa fa-cutlery"></i> Cutlery'),
    ('fa fa-database', '<i class="fa fa-database"></i> Database'),
    ('fa fa-desktop', '<i class="fa fa-desktop"></i> Desktop'),
    ('fa fa-diamond', '<i class="fa fa-diamond"></i> Diamond'),
    ('fa fa-dot-circle-o', '<i class="fa fa-dot-circle-o"></i> Dot-circle-o'),
    ('fa fa-download', '<i class="fa fa-download"></i> Download'),
    ('fa fa-edit', '<i class="fa fa-edit"></i> Edit'),
    ('fa fa-ellipsis-h', '<i class="fa fa-ellipsis-h"></i> Ellipsis-h'),
    ('fa fa-ellipsis-v', '<i class="fa fa-ellipsis-v"></i> Ellipsis-v'),
    ('fa fa-envelope', '<i class="fa fa-envelope"></i> Envelope'),
    ('fa fa-envelope-o', '<i class="fa fa-envelope-o"></i> Envelope-o'),
    ('fa fa-envelope-square', '<i class="fa fa-envelope-square"></i> Envelope-square'),
    ('fa fa-eraser', '<i class="fa fa-eraser"></i> Eraser'),
    ('fa fa-exchange', '<i class="fa fa-exchange"></i> Exchange'),
    ('fa fa-exclamation', '<i class="fa fa-exclamation"></i> Exclamation'),
    ('fa fa-exclamation-circle', '<i class="fa fa-exclamation-circle"></i> Exclamation-circle'),
    ('fa fa-exclamation-triangle', '<i class="fa fa-exclamation-triangle"></i> Exclamation-triangle'),
    ('fa fa-external-link', '<i class="fa fa-external-link"></i> External-link'),
    ('fa fa-external-link-square', '<i class="fa fa-external-link-square"></i> External-link-square'),
    ('fa fa-eye', '<i class="fa fa-eye"></i> Eye'),
    ('fa fa-eye-slash', '<i class="fa fa-eye-slash"></i> Eye-slash'),
    ('fa fa-eyedropper', '<i class="fa fa-eyedropper"></i> Eyedropper'),
    ('fa fa-fax', '<i class="fa fa-fax"></i> Fax'),
    ('fa fa-feed', '<i class="fa fa-feed"></i> Feed'),
    ('fa fa-female', '<i class="fa fa-female"></i> Female'),
    ('fa fa-fighter-jet', '<i class="fa fa-fighter-jet"></i> Fighter-jet'),
    ('fa fa-file-archive-o', '<i class="fa fa-file-archive-o"></i> File-archive-o'),
    ('fa fa-file-audio-o', '<i class="fa fa-file-audio-o"></i> File-audio-o'),
    ('fa fa-file-code-o', '<i class="fa fa-file-code-o"></i> File-code-o'),
    ('fa fa-file-excel-o', '<i class="fa fa-file-excel-o"></i> File-excel-o'),
    ('fa fa-file-image-o', '<i class="fa fa-file-image-o"></i> File-image-o'),
    ('fa fa-file-movie-o', '<i class="fa fa-file-movie-o"></i> File-movie-o'),
    ('fa fa-file-pdf-o', '<i class="fa fa-file-pdf-o"></i> File-pdf-o'),
    ('fa fa-file-photo-o', '<i class="fa fa-file-photo-o"></i> File-photo-o'),
    ('fa fa-file-picture-o', '<i class="fa fa-file-picture-o"></i> File-picture-o'),
    ('fa fa-file-powerpoint-o', '<i class="fa fa-file-powerpoint-o"></i> File-powerpoint-o'),
    ('fa fa-file-sound-o', '<i class="fa fa-file-sound-o"></i> File-sound-o'),
    ('fa fa-file-video-o', '<i class="fa fa-file-video-o"></i> File-video-o'),
    ('fa fa-file-word-o', '<i class="fa fa-file-word-o"></i> File-word-o'),
    ('fa fa-file-zip-o', '<i class="fa fa-file-zip-o"></i> File-zip-o'),
    ('fa fa-film', '<i class="fa fa-film"></i> Film'),
    ('fa fa-filter', '<i class="fa fa-filter"></i> Filter'),
    ('fa fa-fire', '<i class="fa fa-fire"></i> Fire'),
    ('fa fa-fire-extinguisher', '<i class="fa fa-fire-extinguisher"></i> Fire-extinguisher'),
    ('fa fa-flag', '<i class="fa fa-flag"></i> Flag'),
    ('fa fa-flag-checkered', '<i class="fa fa-flag-checkered"></i> Flag-checkered'),
    ('fa fa-flag-o', '<i class="fa fa-flag-o"></i> Flag-o'),
    ('fa fa-flash', '<i class="fa fa-flash"></i> Flash'),
    ('fa fa-flask', '<i class="fa fa-flask"></i> Flask'),
    ('fa fa-folder', '<i class="fa fa-folder"></i> Folder'),
    ('fa fa-folder-o', '<i class="fa fa-folder-o"></i> Folder-o'),
    ('fa fa-folder-open', '<i class="fa fa-folder-open"></i> Folder-open'),
    ('fa fa-folder-open-o', '<i class="fa fa-folder-open-o"></i> Folder-open-o'),
    ('fa fa-frown-o', '<i class="fa fa-frown-o"></i> Frown-o'),
    ('fa fa-futbol-o', '<i class="fa fa-futbol-o"></i> Futbol-o'),
    ('fa fa-gamepad', '<i class="fa fa-gamepad"></i> Gamepad'),
    ('fa fa-gavel', '<i class="fa fa-gavel"></i> Gavel'),
    ('fa fa-gear', '<i class="fa fa-gear"></i> Gear'),
    ('fa fa-gears', '<i class="fa fa-gears"></i> Gears'),
    ('fa fa-gift', '<i class="fa fa-gift"></i> Gift'),
    ('fa fa-glass', '<i class="fa fa-glass"></i> Glass'),
    ('fa fa-globe', '<i class="fa fa-globe"></i> Globe'),
    ('fa fa-graduation-cap', '<i class="fa fa-graduation-cap"></i> Graduation-cap'),
    ('fa fa-group', '<i class="fa fa-group"></i> Group'),
    ('fa fa-hdd-o', '<i class="fa fa-hdd-o"></i> Hdd-o'),
    ('fa fa-headphones', '<i class="fa fa-headphones"></i> Headphones'),
    ('fa fa-heart', '<i class="fa fa-heart"></i> Heart'),
    ('fa fa-heart-o', '<i class="fa fa-heart-o"></i> Heart-o'),
    ('fa fa-heartbeat', '<i class="fa fa-heartbeat"></i> Heartbeat'),
    ('fa fa-history', '<i class="fa fa-history"></i> History'),
    ('fa fa-home', '<i class="fa fa-home"></i> Home'),
    ('fa fa-hotel', '<i class="fa fa-hotel"></i> Hotel'),
    ('fa fa-image', '<i class="fa fa-image"></i> Image'),
    ('fa fa-inbox', '<i class="fa fa-inbox"></i> Inbox'),
    ('fa fa-info', '<i class="fa fa-info"></i> Info'),
    ('fa fa-info-circle', '<i class="fa fa-info-circle"></i> Info-circle'),
    ('fa fa-institution', '<i class="fa fa-institution"></i> Institution'),
    ('fa fa-key', '<i class="fa fa-key"></i> Key'),
    ('fa fa-keyboard-o', '<i class="fa fa-keyboard-o"></i> Keyboard-o'),
    ('fa fa-language', '<i class="fa fa-language"></i> Language'),
    ('fa fa-laptop', '<i class="fa fa-laptop"></i> Laptop'),
    ('fa fa-leaf', '<i class="fa fa-leaf"></i> Leaf'),
    ('fa fa-legal', '<i class="fa fa-legal"></i> Legal'),
    ('fa fa-lemon-o', '<i class="fa fa-lemon-o"></i> Lemon-o'),
    ('fa fa-level-down', '<i class="fa fa-level-down"></i> Level-down'),
    ('fa fa-level-up', '<i class="fa fa-level-up"></i> Level-up'),
    ('fa fa-life-bouy', '<i class="fa fa-life-bouy"></i> Life-bouy'),
    ('fa fa-life-buoy', '<i class="fa fa-life-buoy"></i> Life-buoy'),
    ('fa fa-life-ring', '<i class="fa fa-life-ring"></i> Life-ring'),
    ('fa fa-life-saver', '<i class="fa fa-life-saver"></i> Life-saver'),
    ('fa fa-lightbulb-o', '<i class="fa fa-lightbulb-o"></i> Lightbulb-o'),
    ('fa fa-line-chart', '<i class="fa fa-line-chart"></i> Line-chart'),
    ('fa fa-location-arrow', '<i class="fa fa-location-arrow"></i> Location-arrow'),
    ('fa fa-lock', '<i class="fa fa-lock"></i> Lock'),
    ('fa fa-magic', '<i class="fa fa-magic"></i> Magic'),
    ('fa fa-magnet', '<i class="fa fa-magnet"></i> Magnet'),
    ('fa fa-mail-forward', '<i class="fa fa-mail-forward"></i> Mail-forward'),
    ('fa fa-mail-reply', '<i class="fa fa-mail-reply"></i> Mail-reply'),
    ('fa fa-mail-reply-all', '<i class="fa fa-mail-reply-all"></i> Mail-reply-all'),
    ('fa fa-male', '<i class="fa fa-male"></i> Male'),
    ('fa fa-map-marker', '<i class="fa fa-map-marker"></i> Map-marker'),
    ('fa fa-meh-o', '<i class="fa fa-meh-o"></i> Meh-o'),
    ('fa fa-microphone', '<i class="fa fa-microphone"></i> Microphone'),
    ('fa fa-microphone-slash', '<i class="fa fa-microphone-slash"></i> Microphone-slash'),
    ('fa fa-minus', '<i class="fa fa-minus"></i> Minus'),
    ('fa fa-minus-circle', '<i class="fa fa-minus-circle"></i> Minus-circle'),
    ('fa fa-minus-square', '<i class="fa fa-minus-square"></i> Minus-square'),
    ('fa fa-minus-square-o', '<i class="fa fa-minus-square-o"></i> Minus-square-o'),
    ('fa fa-mobile', '<i class="fa fa-mobile"></i> Mobile'),
    ('fa fa-mobile-phone', '<i class="fa fa-mobile-phone"></i> Mobile-phone'),
    ('fa fa-money', '<i class="fa fa-money"></i> Money'),
    ('fa fa-moon-o', '<i class="fa fa-moon-o"></i> Moon-o'),
    ('fa fa-mortar-board', '<i class="fa fa-mortar-board"></i> Mortar-board'),
    ('fa fa-motorcycle', '<i class="fa fa-motorcycle"></i> Motorcycle'),
    ('fa fa-music', '<i class="fa fa-music"></i> Music'),
    ('fa fa-navicon', '<i class="fa fa-navicon"></i> Navicon'),
    ('fa fa-newspaper-o', '<i class="fa fa-newspaper-o"></i> Newspaper-o'),
    ('fa fa-paint-brush', '<i class="fa fa-paint-brush"></i> Paint-brush'),
    ('fa fa-paper-plane', '<i class="fa fa-paper-plane"></i> Paper-plane'),
    ('fa fa-paper-plane-o', '<i class="fa fa-paper-plane-o"></i> Paper-plane-o'),
    ('fa fa-paw', '<i class="fa fa-paw"></i> Paw'),
    ('fa fa-pencil', '<i class="fa fa-pencil"></i> Pencil'),
    ('fa fa-pencil-square', '<i class="fa fa-pencil-square"></i> Pencil-square'),
    ('fa fa-pencil-square-o', '<i class="fa fa-pencil-square-o"></i> Pencil-square-o'),
    ('fa fa-phone', '<i class="fa fa-phone"></i> Phone'),
    ('fa fa-phone-square', '<i class="fa fa-phone-square"></i> Phone-square'),
    ('fa fa-photo', '<i class="fa fa-photo"></i> Photo'),
    ('fa fa-picture-o', '<i class="fa fa-picture-o"></i> Picture-o'),
    ('fa fa-pie-chart', '<i class="fa fa-pie-chart"></i> Pie-chart'),
    ('fa fa-plane', '<i class="fa fa-plane"></i> Plane'),
    ('fa fa-plug', '<i class="fa fa-plug"></i> Plug'),
    ('fa fa-plus', '<i class="fa fa-plus"></i> Plus'),
    ('fa fa-plus-circle', '<i class="fa fa-plus-circle"></i> Plus-circle'),
    ('fa fa-plus-square', '<i class="fa fa-plus-square"></i> Plus-square'),
    ('fa fa-plus-square-o', '<i class="fa fa-plus-square-o"></i> Plus-square-o'),
    ('fa fa-power-off', '<i class="fa fa-power-off"></i> Power-off'),
    ('fa fa-print', '<i class="fa fa-print"></i> Print'),
    ('fa fa-puzzle-piece', '<i class="fa fa-puzzle-piece"></i> Puzzle-piece'),
    ('fa fa-qrcode', '<i class="fa fa-qrcode"></i> Qrcode'),
    ('fa fa-question', '<i class="fa fa-question"></i> Question'),
    ('fa fa-question-circle', '<i class="fa fa-question-circle"></i> Question-circle'),
    ('fa fa-quote-left', '<i class="fa fa-quote-left"></i> Quote-left'),
    ('fa fa-quote-right', '<i class="fa fa-quote-right"></i> Quote-right'),
    ('fa fa-random', '<i class="fa fa-random"></i> Random'),
    ('fa fa-recycle', '<i class="fa fa-recycle"></i> Recycle'),
    ('fa fa-refresh', '<i class="fa fa-refresh"></i> Refresh'),
    ('fa fa-remove', '<i class="fa fa-remove"></i> Remove'),
    ('fa fa-reorder', '<i class="fa fa-reorder"></i> Reorder'),
    ('fa fa-reply', '<i class="fa fa-reply"></i> Reply'),
    ('fa fa-reply-all', '<i class="fa fa-reply-all"></i> Reply-all'),
    ('fa fa-retweet', '<i class="fa fa-retweet"></i> Retweet'),
    ('fa fa-road', '<i class="fa fa-road"></i> Road'),
    ('fa fa-rocket', '<i class="fa fa-rocket"></i> Rocket'),
    ('fa fa-rss', '<i class="fa fa-rss"></i> Rss'),
    ('fa fa-rss-square', '<i class="fa fa-rss-square"></i> Rss-square'),
    ('fa fa-search', '<i class="fa fa-search"></i> Search'),
    ('fa fa-search-minus', '<i class="fa fa-search-minus"></i> Search-minus'),
    ('fa fa-search-plus', '<i class="fa fa-search-plus"></i> Search-plus'),
    ('fa fa-send', '<i class="fa fa-send"></i> Send'),
    ('fa fa-send-o', '<i class="fa fa-send-o"></i> Send-o'),
    ('fa fa-server', '<i class="fa fa-server"></i> Server'),
    ('fa fa-share', '<i class="fa fa-share"></i> Share'),
    ('fa fa-share-alt', '<i class="fa fa-share-alt"></i> Share-alt'),
    ('fa fa-share-alt-square', '<i class="fa fa-share-alt-square"></i> Share-alt-square'),
    ('fa fa-share-square', '<i class="fa fa-share-square"></i> Share-square'),
    ('fa fa-share-square-o', '<i class="fa fa-share-square-o"></i> Share-square-o'),
    ('fa fa-shield', '<i class="fa fa-shield"></i> Shield'),
    ('fa fa-ship', '<i class="fa fa-ship"></i> Ship'),
    ('fa fa-shopping-cart', '<i class="fa fa-shopping-cart"></i> Shopping-cart'),
    ('fa fa-sign-in', '<i class="fa fa-sign-in"></i> Sign-in'),
    ('fa fa-sign-out', '<i class="fa fa-sign-out"></i> Sign-out'),
    ('fa fa-signal', '<i class="fa fa-signal"></i> Signal'),
    ('fa fa-sitemap', '<i class="fa fa-sitemap"></i> Sitemap'),
    ('fa fa-sliders', '<i class="fa fa-sliders"></i> Sliders'),
    ('fa fa-smile-o', '<i class="fa fa-smile-o"></i> Smile-o'),
    ('fa fa-soccer-ball-o', '<i class="fa fa-soccer-ball-o"></i> Soccer-ball-o'),
    ('fa fa-sort', '<i class="fa fa-sort"></i> Sort'),
    ('fa fa-sort-alpha-asc', '<i class="fa fa-sort-alpha-asc"></i> Sort-alpha-asc'),
    ('fa fa-sort-alpha-desc', '<i class="fa fa-sort-alpha-desc"></i> Sort-alpha-desc'),
    ('fa fa-sort-amount-asc', '<i class="fa fa-sort-amount-asc"></i> Sort-amount-asc'),
    ('fa fa-sort-amount-desc', '<i class="fa fa-sort-amount-desc"></i> Sort-amount-desc'),
    ('fa fa-sort-asc', '<i class="fa fa-sort-asc"></i> Sort-asc'),
    ('fa fa-sort-desc', '<i class="fa fa-sort-desc"></i> Sort-desc'),
    ('fa fa-sort-down', '<i class="fa fa-sort-down"></i> Sort-down'),
    ('fa fa-sort-numeric-asc', '<i class="fa fa-sort-numeric-asc"></i> Sort-numeric-asc'),
    ('fa fa-sort-numeric-desc', '<i class="fa fa-sort-numeric-desc"></i> Sort-numeric-desc'),
    ('fa fa-sort-up', '<i class="fa fa-sort-up"></i> Sort-up'),
    ('fa fa-space-shuttle', '<i class="fa fa-space-shuttle"></i> Space-shuttle'),
    ('fa fa-spinner', '<i class="fa fa-spinner"></i> Spinner'),
    ('fa fa-spoon', '<i class="fa fa-spoon"></i> Spoon'),
    ('fa fa-square', '<i class="fa fa-square"></i> Square'),
    ('fa fa-square-o', '<i class="fa fa-square-o"></i> Square-o'),
    ('fa fa-star', '<i class="fa fa-star"></i> Star'),
    ('fa fa-star-half', '<i class="fa fa-star-half"></i> Star-half'),
    ('fa fa-star-half-empty', '<i class="fa fa-star-half-empty"></i> Star-half-empty'),
    ('fa fa-star-half-full', '<i class="fa fa-star-half-full"></i> Star-half-full'),
    ('fa fa-star-half-o', '<i class="fa fa-star-half-o"></i> Star-half-o'),
    ('fa fa-star-o', '<i class="fa fa-star-o"></i> Star-o'),
    ('fa fa-street-view', '<i class="fa fa-street-view"></i> Street-view'),
    ('fa fa-suitcase', '<i class="fa fa-suitcase"></i> Suitcase'),
    ('fa fa-sun-o', '<i class="fa fa-sun-o"></i> Sun-o'),
    ('fa fa-support', '<i class="fa fa-support"></i> Support'),
    ('fa fa-tablet', '<i class="fa fa-tablet"></i> Tablet'),
    ('fa fa-tachometer', '<i class="fa fa-tachometer"></i> Tachometer'),
    ('fa fa-tag', '<i class="fa fa-tag"></i> Tag'),
    ('fa fa-tags', '<i class="fa fa-tags"></i> Tags'),
    ('fa fa-tasks', '<i class="fa fa-tasks"></i> Tasks'),
    ('fa fa-taxi', '<i class="fa fa-taxi"></i> Taxi'),
    ('fa fa-terminal', '<i class="fa fa-terminal"></i> Terminal'),
    ('fa fa-thumb-tack', '<i class="fa fa-thumb-tack"></i> Thumb-tack'),
    ('fa fa-thumbs-down', '<i class="fa fa-thumbs-down"></i> Thumbs-down'),
    ('fa fa-thumbs-o-down', '<i class="fa fa-thumbs-o-down"></i> Thumbs-o-down'),
    ('fa fa-thumbs-o-up', '<i class="fa fa-thumbs-o-up"></i> Thumbs-o-up'),
    ('fa fa-thumbs-up', '<i class="fa fa-thumbs-up"></i> Thumbs-up'),
    ('fa fa-ticket', '<i class="fa fa-ticket"></i> Ticket'),
    ('fa fa-times', '<i class="fa fa-times"></i> Times'),
    ('fa fa-times-circle', '<i class="fa fa-times-circle"></i> Times-circle'),
    ('fa fa-times-circle-o', '<i class="fa fa-times-circle-o"></i> Times-circle-o'),
    ('fa fa-tint', '<i class="fa fa-tint"></i> Tint'),
    ('fa fa-toggle-down', '<i class="fa fa-toggle-down"></i> Toggle-down'),
    ('fa fa-toggle-left', '<i class="fa fa-toggle-left"></i> Toggle-left'),
    ('fa fa-toggle-off', '<i class="fa fa-toggle-off"></i> Toggle-off'),
    ('fa fa-toggle-on', '<i class="fa fa-toggle-on"></i> Toggle-on'),
    ('fa fa-toggle-right', '<i class="fa fa-toggle-right"></i> Toggle-right'),
    ('fa fa-toggle-up', '<i class="fa fa-toggle-up"></i> Toggle-up'),
    ('fa fa-trash', '<i class="fa fa-trash"></i> Trash'),
    ('fa fa-trash-o', '<i class="fa fa-trash-o"></i> Trash-o'),
    ('fa fa-tree', '<i class="fa fa-tree"></i> Tree'),
    ('fa fa-trophy', '<i class="fa fa-trophy"></i> Trophy'),
    ('fa fa-truck', '<i class="fa fa-truck"></i> Truck'),
    ('fa fa-tty', '<i class="fa fa-tty"></i> Tty'),
    ('fa fa-umbrella', '<i class="fa fa-umbrella"></i> Umbrella'),
    ('fa fa-university', '<i class="fa fa-university"></i> University'),
    ('fa fa-unlock', '<i class="fa fa-unlock"></i> Unlock'),
    ('fa fa-unlock-alt', '<i class="fa fa-unlock-alt"></i> Unlock-alt'),
    ('fa fa-unsorted', '<i class="fa fa-unsorted"></i> Unsorted'),
    ('fa fa-upload', '<i class="fa fa-upload"></i> Upload'),
    ('fa fa-user', '<i class="fa fa-user"></i> User'),
    ('fa fa-user-plus', '<i class="fa fa-user-plus"></i> User-plus'),
    ('fa fa-user-secret', '<i class="fa fa-user-secret"></i> User-secret'),
    ('fa fa-user-times', '<i class="fa fa-user-times"></i> User-times'),
    ('fa fa-users', '<i class="fa fa-users"></i> Users'),
    ('fa fa-video-camera', '<i class="fa fa-video-camera"></i> Video-camera'),
    ('fa fa-volume-down', '<i class="fa fa-volume-down"></i> Volume-down'),
    ('fa fa-volume-off', '<i class="fa fa-volume-off"></i> Volume-off'),
    ('fa fa-volume-up', '<i class="fa fa-volume-up"></i> Volume-up'),
    ('fa fa-warning', '<i class="fa fa-warning"></i> Warning'),
    ('fa fa-wheelchair', '<i class="fa fa-wheelchair"></i> Wheelchair'),
    ('fa fa-wifi', '<i class="fa fa-wifi"></i> Wifi'),
    ('fa fa-wrench', '<i class="fa fa-wrench"></i> Wrench'),
    ('fa fa-hand-o-down', '<i class="fa fa-hand-o-down"></i> Hand-o-down'),
    ('fa fa-hand-o-left', '<i class="fa fa-hand-o-left"></i> Hand-o-left'),
    ('fa fa-hand-o-right', '<i class="fa fa-hand-o-right"></i> Hand-o-right'),
    ('fa fa-hand-o-up', '<i class="fa fa-hand-o-up"></i> Hand-o-up'),
    ('fa fa-ambulance', '<i class="fa fa-ambulance"></i> Ambulance'),
    ('fa fa-subway', '<i class="fa fa-subway"></i> Subway'),
    ('fa fa-train', '<i class="fa fa-train"></i> Train'),
    ('fa fa-intersex', '<i class="fa fa-intersex"></i> Intersex'),
    ('fa fa-mars', '<i class="fa fa-mars"></i> Mars'),
    ('fa fa-mars-double', '<i class="fa fa-mars-double"></i> Mars-double'),
    ('fa fa-mars-stroke', '<i class="fa fa-mars-stroke"></i> Mars-stroke'),
    ('fa fa-mars-stroke-h', '<i class="fa fa-mars-stroke-h"></i> Mars-stroke-h'),
    ('fa fa-mars-stroke-v', '<i class="fa fa-mars-stroke-v"></i> Mars-stroke-v'),
    ('fa fa-mercury', '<i class="fa fa-mercury"></i> Mercury'),
    ('fa fa-neuter', '<i class="fa fa-neuter"></i> Neuter'),
    ('fa fa-transgender', '<i class="fa fa-transgender"></i> Transgender'),
    ('fa fa-transgender-alt', '<i class="fa fa-transgender-alt"></i> Transgender-alt'),
    ('fa fa-venus', '<i class="fa fa-venus"></i> Venus'),
    ('fa fa-venus-double', '<i class="fa fa-venus-double"></i> Venus-double'),
    ('fa fa-venus-mars', '<i class="fa fa-venus-mars"></i> Venus-mars'),
    ('fa fa-file', '<i class="fa fa-file"></i> File'),
    ('fa fa-file-o', '<i class="fa fa-file-o"></i> File-o'),
    ('fa fa-file-text', '<i class="fa fa-file-text"></i> File-text'),
    ('fa fa-file-text-o', '<i class="fa fa-file-text-o"></i> File-text-o'),
    ('fa fa-cc-amex', '<i class="fa fa-cc-amex"></i> Cc-amex'),
    ('fa fa-cc-discover', '<i class="fa fa-cc-discover"></i> Cc-discover'),
    ('fa fa-cc-mastercard', '<i class="fa fa-cc-mastercard"></i> Cc-mastercard'),
    ('fa fa-cc-paypal', '<i class="fa fa-cc-paypal"></i> Cc-paypal'),
    ('fa fa-cc-stripe', '<i class="fa fa-cc-stripe"></i> Cc-stripe'),
    ('fa fa-cc-visa', '<i class="fa fa-cc-visa"></i> Cc-visa'),
    ('fa fa-google-wallet', '<i class="fa fa-google-wallet"></i> Google-wallet'),
    ('fa fa-paypal', '<i class="fa fa-paypal"></i> Paypal'),
    ('fa fa-bitcoin', '<i class="fa fa-bitcoin"></i> Bitcoin'),
    ('fa fa-btc', '<i class="fa fa-btc"></i> Btc'),
    ('fa fa-cny', '<i class="fa fa-cny"></i> Cny'),
    ('fa fa-dollar', '<i class="fa fa-dollar"></i> Dollar'),
    ('fa fa-eur', '<i class="fa fa-eur"></i> Eur'),
    ('fa fa-euro', '<i class="fa fa-euro"></i> Euro'),
    ('fa fa-gbp', '<i class="fa fa-gbp"></i> Gbp'),
    ('fa fa-ils', '<i class="fa fa-ils"></i> Ils'),
    ('fa fa-inr', '<i class="fa fa-inr"></i> Inr'),
    ('fa fa-jpy', '<i class="fa fa-jpy"></i> Jpy'),
    ('fa fa-krw', '<i class="fa fa-krw"></i> Krw'),
    ('fa fa-rmb', '<i class="fa fa-rmb"></i> Rmb'),
    ('fa fa-rouble', '<i class="fa fa-rouble"></i> Rouble'),
    ('fa fa-rub', '<i class="fa fa-rub"></i> Rub'),
    ('fa fa-ruble', '<i class="fa fa-ruble"></i> Ruble'),
    ('fa fa-rupee', '<i class="fa fa-rupee"></i> Rupee'),
    ('fa fa-shekel', '<i class="fa fa-shekel"></i> Shekel'),
    ('fa fa-sheqel', '<i class="fa fa-sheqel"></i> Sheqel'),
    ('fa fa-try', '<i class="fa fa-try"></i> Try'),
    ('fa fa-turkish-lira', '<i class="fa fa-turkish-lira"></i> Turkish-lira'),
    ('fa fa-usd', '<i class="fa fa-usd"></i> Usd'),
    ('fa fa-won', '<i class="fa fa-won"></i> Won'),
    ('fa fa-yen', '<i class="fa fa-yen"></i> Yen'),
    ('fa fa-align-center', '<i class="fa fa-align-center"></i> Align-center'),
    ('fa fa-align-justify', '<i class="fa fa-align-justify"></i> Align-justify'),
    ('fa fa-align-left', '<i class="fa fa-align-left"></i> Align-left'),
    ('fa fa-align-right', '<i class="fa fa-align-right"></i> Align-right'),
    ('fa fa-bold', '<i class="fa fa-bold"></i> Bold'),
    ('fa fa-chain', '<i class="fa fa-chain"></i> Chain'),
    ('fa fa-chain-broken', '<i class="fa fa-chain-broken"></i> Chain-broken'),
    ('fa fa-clipboard', '<i class="fa fa-clipboard"></i> Clipboard'),
    ('fa fa-columns', '<i class="fa fa-columns"></i> Columns'),
    ('fa fa-copy', '<i class="fa fa-copy"></i> Copy'),
    ('fa fa-cut', '<i class="fa fa-cut"></i> Cut'),
    ('fa fa-dedent', '<i class="fa fa-dedent"></i> Dedent'),
    ('fa fa-files-o', '<i class="fa fa-files-o"></i> Files-o'),
    ('fa fa-floppy-o', '<i class="fa fa-floppy-o"></i> Floppy-o'),
    ('fa fa-font', '<i class="fa fa-font"></i> Font'),
    ('fa fa-header', '<i class="fa fa-header"></i> Header'),
    ('fa fa-indent', '<i class="fa fa-indent"></i> Indent'),
    ('fa fa-italic', '<i class="fa fa-italic"></i> Italic'),
    ('fa fa-link', '<i class="fa fa-link"></i> Link'),
    ('fa fa-list', '<i class="fa fa-list"></i> List'),
    ('fa fa-list-alt', '<i class="fa fa-list-alt"></i> List-alt'),
    ('fa fa-list-ol', '<i class="fa fa-list-ol"></i> List-ol'),
    ('fa fa-list-ul', '<i class="fa fa-list-ul"></i> List-ul'),
    ('fa fa-outdent', '<i class="fa fa-outdent"></i> Outdent'),
    ('fa fa-paperclip', '<i class="fa fa-paperclip"></i> Paperclip'),
    ('fa fa-paragraph', '<i class="fa fa-paragraph"></i> Paragraph'),
    ('fa fa-paste', '<i class="fa fa-paste"></i> Paste'),
    ('fa fa-repeat', '<i class="fa fa-repeat"></i> Repeat'),
    ('fa fa-rotate-left', '<i class="fa fa-rotate-left"></i> Rotate-left'),
    ('fa fa-rotate-right', '<i class="fa fa-rotate-right"></i> Rotate-right'),
    ('fa fa-save', '<i class="fa fa-save"></i> Save'),
    ('fa fa-scissors', '<i class="fa fa-scissors"></i> Scissors'),
    ('fa fa-strikethrough', '<i class="fa fa-strikethrough"></i> Strikethrough'),
    ('fa fa-subscript', '<i class="fa fa-subscript"></i> Subscript'),
    ('fa fa-superscript', '<i class="fa fa-superscript"></i> Superscript'),
    ('fa fa-table', '<i class="fa fa-table"></i> Table'),
    ('fa fa-text-height', '<i class="fa fa-text-height"></i> Text-height'),
    ('fa fa-text-width', '<i class="fa fa-text-width"></i> Text-width'),
    ('fa fa-th', '<i class="fa fa-th"></i> Th'),
    ('fa fa-th-large', '<i class="fa fa-th-large"></i> Th-large'),
    ('fa fa-th-list', '<i class="fa fa-th-list"></i> Th-list'),
    ('fa fa-underline', '<i class="fa fa-underline"></i> Underline'),
    ('fa fa-undo', '<i class="fa fa-undo"></i> Undo'),
    ('fa fa-unlink', '<i class="fa fa-unlink"></i> Unlink'),
    ('fa fa-angle-double-down', '<i class="fa fa-angle-double-down"></i> Angle-double-down'),
    ('fa fa-angle-double-left', '<i class="fa fa-angle-double-left"></i> Angle-double-left'),
    ('fa fa-angle-double-right', '<i class="fa fa-angle-double-right"></i> Angle-double-right'),
    ('fa fa-angle-double-up', '<i class="fa fa-angle-double-up"></i> Angle-double-up'),
    ('fa fa-angle-down', '<i class="fa fa-angle-down"></i> Angle-down'),
    ('fa fa-angle-left', '<i class="fa fa-angle-left"></i> Angle-left'),
    ('fa fa-angle-right', '<i class="fa fa-angle-right"></i> Angle-right'),
    ('fa fa-angle-up', '<i class="fa fa-angle-up"></i> Angle-up'),
    ('fa fa-arrow-circle-down', '<i class="fa fa-arrow-circle-down"></i> Arrow-circle-down'),
    ('fa fa-arrow-circle-left', '<i class="fa fa-arrow-circle-left"></i> Arrow-circle-left'),
    ('fa fa-arrow-circle-o-down', '<i class="fa fa-arrow-circle-o-down"></i> Arrow-circle-o-down'),
    ('fa fa-arrow-circle-o-left', '<i class="fa fa-arrow-circle-o-left"></i> Arrow-circle-o-left'),
    ('fa fa-arrow-circle-o-right', '<i class="fa fa-arrow-circle-o-right"></i> Arrow-circle-o-right'),
    ('fa fa-arrow-circle-o-up', '<i class="fa fa-arrow-circle-o-up"></i> Arrow-circle-o-up'),
    ('fa fa-arrow-circle-right', '<i class="fa fa-arrow-circle-right"></i> Arrow-circle-right'),
    ('fa fa-arrow-circle-up', '<i class="fa fa-arrow-circle-up"></i> Arrow-circle-up'),
    ('fa fa-arrow-down', '<i class="fa fa-arrow-down"></i> Arrow-down'),
    ('fa fa-arrow-left', '<i class="fa fa-arrow-left"></i> Arrow-left'),
    ('fa fa-arrow-right', '<i class="fa fa-arrow-right"></i> Arrow-right'),
    ('fa fa-arrow-up', '<i class="fa fa-arrow-up"></i> Arrow-up'),
    ('fa fa-arrows-alt', '<i class="fa fa-arrows-alt"></i> Arrows-alt'),
    ('fa fa-caret-down', '<i class="fa fa-caret-down"></i> Caret-down'),
    ('fa fa-caret-left', '<i class="fa fa-caret-left"></i> Caret-left'),
    ('fa fa-caret-right', '<i class="fa fa-caret-right"></i> Caret-right'),
    ('fa fa-caret-up', '<i class="fa fa-caret-up"></i> Caret-up'),
    ('fa fa-chevron-circle-down', '<i class="fa fa-chevron-circle-down"></i> Chevron-circle-down'),
    ('fa fa-chevron-circle-left', '<i class="fa fa-chevron-circle-left"></i> Chevron-circle-left'),
    ('fa fa-chevron-circle-right', '<i class="fa fa-chevron-circle-right"></i> Chevron-circle-right'),
    ('fa fa-chevron-circle-up', '<i class="fa fa-chevron-circle-up"></i> Chevron-circle-up'),
    ('fa fa-chevron-down', '<i class="fa fa-chevron-down"></i> Chevron-down'),
    ('fa fa-chevron-left', '<i class="fa fa-chevron-left"></i> Chevron-left'),
    ('fa fa-chevron-right', '<i class="fa fa-chevron-right"></i> Chevron-right'),
    ('fa fa-chevron-up', '<i class="fa fa-chevron-up"></i> Chevron-up'),
    ('fa fa-long-arrow-down', '<i class="fa fa-long-arrow-down"></i> Long-arrow-down'),
    ('fa fa-long-arrow-left', '<i class="fa fa-long-arrow-left"></i> Long-arrow-left'),
    ('fa fa-long-arrow-right', '<i class="fa fa-long-arrow-right"></i> Long-arrow-right'),
    ('fa fa-long-arrow-up', '<i class="fa fa-long-arrow-up"></i> Long-arrow-up'),
    ('fa fa-backward', '<i class="fa fa-backward"></i> Backward'),
    ('fa fa-compress', '<i class="fa fa-compress"></i> Compress'),
    ('fa fa-eject', '<i class="fa fa-eject"></i> Eject'),
    ('fa fa-expand', '<i class="fa fa-expand"></i> Expand'),
    ('fa fa-fast-backward', '<i class="fa fa-fast-backward"></i> Fast-backward'),
    ('fa fa-fast-forward', '<i class="fa fa-fast-forward"></i> Fast-forward'),
    ('fa fa-forward', '<i class="fa fa-forward"></i> Forward'),
    ('fa fa-pause', '<i class="fa fa-pause"></i> Pause'),
    ('fa fa-play', '<i class="fa fa-play"></i> Play'),
    ('fa fa-play-circle', '<i class="fa fa-play-circle"></i> Play-circle'),
    ('fa fa-play-circle-o', '<i class="fa fa-play-circle-o"></i> Play-circle-o'),
    ('fa fa-step-backward', '<i class="fa fa-step-backward"></i> Step-backward'),
    ('fa fa-step-forward', '<i class="fa fa-step-forward"></i> Step-forward'),
    ('fa fa-stop', '<i class="fa fa-stop"></i> Stop'),
    ('fa fa-youtube-play', '<i class="fa fa-youtube-play"></i> Youtube-play'),
    ('fa fa-adn', '<i class="fa fa-adn"></i> Adn'),
    ('fa fa-android', '<i class="fa fa-android"></i> Android'),
    ('fa fa-angellist', '<i class="fa fa-angellist"></i> Angellist'),
    ('fa fa-apple', '<i class="fa fa-apple"></i> Apple'),
    ('fa fa-behance', '<i class="fa fa-behance"></i> Behance'),
    ('fa fa-behance-square', '<i class="fa fa-behance-square"></i> Behance-square'),
    ('fa fa-bitbucket', '<i class="fa fa-bitbucket"></i> Bitbucket'),
    ('fa fa-bitbucket-square', '<i class="fa fa-bitbucket-square"></i> Bitbucket-square'),
    ('fa fa-buysellads', '<i class="fa fa-buysellads"></i> Buysellads'),
    ('fa fa-codepen', '<i class="fa fa-codepen"></i> Codepen'),
    ('fa fa-connectdevelop', '<i class="fa fa-connectdevelop"></i> Connectdevelop'),
    ('fa fa-css3', '<i class="fa fa-css3"></i> Css3'),
    ('fa fa-dashcube', '<i class="fa fa-dashcube"></i> Dashcube'),
    ('fa fa-delicious', '<i class="fa fa-delicious"></i> Delicious'),
    ('fa fa-deviantart', '<i class="fa fa-deviantart"></i> Deviantart'),
    ('fa fa-digg', '<i class="fa fa-digg"></i> Digg'),
    ('fa fa-dribbble', '<i class="fa fa-dribbble"></i> Dribbble'),
    ('fa fa-dropbox', '<i class="fa fa-dropbox"></i> Dropbox'),
    ('fa fa-drupal', '<i class="fa fa-drupal"></i> Drupal'),
    ('fa fa-empire', '<i class="fa fa-empire"></i> Empire'),
    ('fa fa-facebook', '<i class="fa fa-facebook"></i> Facebook'),
    ('fa fa-facebook-f', '<i class="fa fa-facebook-f"></i> Facebook-f'),
    ('fa fa-facebook-official', '<i class="fa fa-facebook-official"></i> Facebook-official'),
    ('fa fa-facebook-square', '<i class="fa fa-facebook-square"></i> Facebook-square'),
    ('fa fa-flickr', '<i class="fa fa-flickr"></i> Flickr'),
    ('fa fa-forumbee', '<i class="fa fa-forumbee"></i> Forumbee'),
    ('fa fa-foursquare', '<i class="fa fa-foursquare"></i> Foursquare'),
    ('fa fa-ge', '<i class="fa fa-ge"></i> Ge'),
    ('fa fa-git', '<i class="fa fa-git"></i> Git'),
    ('fa fa-git-square', '<i class="fa fa-git-square"></i> Git-square'),
    ('fa fa-github', '<i class="fa fa-github"></i> Github'),
    ('fa fa-github-alt', '<i class="fa fa-github-alt"></i> Github-alt'),
    ('fa fa-github-square', '<i class="fa fa-github-square"></i> Github-square'),
    ('fa fa-gittip', '<i class="fa fa-gittip"></i> Gittip'),
    ('fa fa-google', '<i class="fa fa-google"></i> Google'),
    ('fa fa-google-plus', '<i class="fa fa-google-plus"></i> Google-plus'),
    ('fa fa-google-plus-square', '<i class="fa fa-google-plus-square"></i> Google-plus-square'),
    ('fa fa-gratipay', '<i class="fa fa-gratipay"></i> Gratipay'),
    ('fa fa-hacker-news', '<i class="fa fa-hacker-news"></i> Hacker-news'),
    ('fa fa-html5', '<i class="fa fa-html5"></i> Html5'),
    ('fa fa-instagram', '<i class="fa fa-instagram"></i> Instagram'),
    ('fa fa-ioxhost', '<i class="fa fa-ioxhost"></i> Ioxhost'),
    ('fa fa-joomla', '<i class="fa fa-joomla"></i> Joomla'),
    ('fa fa-jsfiddle', '<i class="fa fa-jsfiddle"></i> Jsfiddle'),
    ('fa fa-lastfm', '<i class="fa fa-lastfm"></i> Lastfm'),
    ('fa fa-lastfm-square', '<i class="fa fa-lastfm-square"></i> Lastfm-square'),
    ('fa fa-leanpub', '<i class="fa fa-leanpub"></i> Leanpub'),
    ('fa fa-linkedin', '<i class="fa fa-linkedin"></i> Linkedin'),
    ('fa fa-linkedin-square', '<i class="fa fa-linkedin-square"></i> Linkedin-square'),
    ('fa fa-linux', '<i class="fa fa-linux"></i> Linux'),
    ('fa fa-maxcdn', '<i class="fa fa-maxcdn"></i> Maxcdn'),
    ('fa fa-meanpath', '<i class="fa fa-meanpath"></i> Meanpath'),
    ('fa fa-medium', '<i class="fa fa-medium"></i> Medium'),
    ('fa fa-openid', '<i class="fa fa-openid"></i> Openid'),
    ('fa fa-pagelines', '<i class="fa fa-pagelines"></i> Pagelines'),
    ('fa fa-pied-piper', '<i class="fa fa-pied-piper"></i> Pied-piper'),
    ('fa fa-pied-piper-alt', '<i class="fa fa-pied-piper-alt"></i> Pied-piper-alt'),
    ('fa fa-pinterest', '<i class="fa fa-pinterest"></i> Pinterest'),
    ('fa fa-pinterest-p', '<i class="fa fa-pinterest-p"></i> Pinterest-p'),
    ('fa fa-pinterest-square', '<i class="fa fa-pinterest-square"></i> Pinterest-square'),
    ('fa fa-qq', '<i class="fa fa-qq"></i> Qq'),
    ('fa fa-ra', '<i class="fa fa-ra"></i> Ra'),
    ('fa fa-rebel', '<i class="fa fa-rebel"></i> Rebel'),
    ('fa fa-reddit', '<i class="fa fa-reddit"></i> Reddit'),
    ('fa fa-reddit-square', '<i class="fa fa-reddit-square"></i> Reddit-square'),
    ('fa fa-renren', '<i class="fa fa-renren"></i> Renren'),
    ('fa fa-sellsy', '<i class="fa fa-sellsy"></i> Sellsy'),
    ('fa fa-shirtsinbulk', '<i class="fa fa-shirtsinbulk"></i> Shirtsinbulk'),
    ('fa fa-simplybuilt', '<i class="fa fa-simplybuilt"></i> Simplybuilt'),
    ('fa fa-skyatlas', '<i class="fa fa-skyatlas"></i> Skyatlas'),
    ('fa fa-skype', '<i class="fa fa-skype"></i> Skype'),
    ('fa fa-slack', '<i class="fa fa-slack"></i> Slack'),
    ('fa fa-slideshare', '<i class="fa fa-slideshare"></i> Slideshare'),
    ('fa fa-soundcloud', '<i class="fa fa-soundcloud"></i> Soundcloud'),
    ('fa fa-spotify', '<i class="fa fa-spotify"></i> Spotify'),
    ('fa fa-stack-exchange', '<i class="fa fa-stack-exchange"></i> Stack-exchange'),
    ('fa fa-stack-overflow', '<i class="fa fa-stack-overflow"></i> Stack-overflow'),
    ('fa fa-steam', '<i class="fa fa-steam"></i> Steam'),
    ('fa fa-steam-square', '<i class="fa fa-steam-square"></i> Steam-square'),
    ('fa fa-stumbleupon', '<i class="fa fa-stumbleupon"></i> Stumbleupon'),
    ('fa fa-stumbleupon-circle', '<i class="fa fa-stumbleupon-circle"></i> Stumbleupon-circle'),
    ('fa fa-tencent-weibo', '<i class="fa fa-tencent-weibo"></i> Tencent-weibo'),
    ('fa fa-trello', '<i class="fa fa-trello"></i> Trello'),
    ('fa fa-tumblr', '<i class="fa fa-tumblr"></i> Tumblr'),
    ('fa fa-tumblr-square', '<i class="fa fa-tumblr-square"></i> Tumblr-square'),
    ('fa fa-twitch', '<i class="fa fa-twitch"></i> Twitch'),
    ('fa fa-twitter', '<i class="fa fa-twitter"></i> Twitter'),
    ('fa fa-twitter-square', '<i class="fa fa-twitter-square"></i> Twitter-square'),
    ('fa fa-viacoin', '<i class="fa fa-viacoin"></i> Viacoin'),
    ('fa fa-vimeo-square', '<i class="fa fa-vimeo-square"></i> Vimeo-square'),
    ('fa fa-vine', '<i class="fa fa-vine"></i> Vine'),
    ('fa fa-vk', '<i class="fa fa-vk"></i> Vk'),
    ('fa fa-wechat', '<i class="fa fa-wechat"></i> Wechat'),
    ('fa fa-weibo', '<i class="fa fa-weibo"></i> Weibo'),
    ('fa fa-weixin', '<i class="fa fa-weixin"></i> Weixin'),
    ('fa fa-whatsapp', '<i class="fa fa-whatsapp"></i> Whatsapp'),
    ('fa fa-windows', '<i class="fa fa-windows"></i> Windows'),
    ('fa fa-wordpress', '<i class="fa fa-wordpress"></i> Wordpress'),
    ('fa fa-xing', '<i class="fa fa-xing"></i> Xing'),
    ('fa fa-xing-square', '<i class="fa fa-xing-square"></i> Xing-square'),
    ('fa fa-y-combinator-square', '<i class="fa fa-y-combinator-square"></i> Y-combinator-square'),
    ('fa fa-yahoo', '<i class="fa fa-yahoo"></i> Yahoo'),
    ('fa fa-yc-square', '<i class="fa fa-yc-square"></i> Yc-square'),
    ('fa fa-yelp', '<i class="fa fa-yelp"></i> Yelp'),
    ('fa fa-youtube', '<i class="fa fa-youtube"></i> Youtube'),
    ('fa fa-youtube-square', '<i class="fa fa-youtube-square"></i> Youtube-square'),
    ('fa fa-h-square', '<i class="fa fa-h-square"></i> H-square'),
    ('fa fa-hospital-o', '<i class="fa fa-hospital-o"></i> Hospital-o'),
    ('fa fa-medkit', '<i class="fa fa-medkit"></i> Medkit'),
    ('fa fa-stethoscope', '<i class="fa fa-stethoscope"></i> Stethoscope'),
    ('fa fa-user-md', '<i class="fa fa-user-md"></i> User-md'),
]



"""
trionyx.forms
~~~~~~~~~~~~~

Core forms for Trionyx

:copyright: 2018 by Maikel Martens
:license: GPLv3
"""
import logging
import inspect
from collections import defaultdict

from django.forms import *  # noqa F403
from django.forms import ModelForm as DjangoModelForm
from django.db.models import NOT_PROVIDED
from django.db import transaction

from trionyx.config import models_config

logger = logging.getLogger(__name__)


class ModelForm(DjangoModelForm):
    """Trionyx ModelForm"""

    css_files = []
    """CSS files that need to be loaded"""

    js_files = []
    """JS files that need to be loaded"""

    def get_inline_forms(self):
        """Get inline forms"""
        if not hasattr(self, '__inline_forms'):
            self.__inline_forms = {}
            for key, options in getattr(self, 'inline_forms', {}).items():
                kwargs = {
                    'prefix': key,
                }

                fk_name = options.get('fk_name', 'instance')
                if fk_name == 'instance':
                    kwargs['instance'] = self.instance
                else:
                    kwargs['instance'] = getattr(self.instance, key, None)
                    kwargs['initial'] = {
                        fk_name: self.instance
                    }

                self.__inline_forms[key] = options['form'](self.data if self.data else None, **kwargs)

        return self.__inline_forms

    def is_valid(self):
        """Check if form and inline forms are valid"""
        valid = super().is_valid()

        for key, form in self.get_inline_forms().items():
            valid = valid and form.is_valid()

        return valid

    def save(self, commit=True):
        """Save form and inline forms"""
        object_updated = False
        config = models_config.get_config(self._meta.model)
        fields = [field.name for field in config.get_fields()]
        with transaction.atomic():
            obj = super().save(commit)

            for key, form in self.get_inline_forms().items():
                fk_name = self.inline_forms[key].get('fk_name', 'instance')
                if fk_name == 'instance':
                    logger.debug('Save inline form {} as FormSet, commit: {}'.format(key, commit))
                    form.instance = obj
                    form.save(commit)
                else:
                    logger.debug('Save inline form {} as Form, commit: {}'.format(key, commit))
                    inline_obj = form.save(False)
                    if inline_obj:
                        setattr(inline_obj, fk_name, obj)
                        if commit:
                            inline_obj.save()

                        if key in fields:
                            logger.debug(' * Set inline form object on parent form as: {}'.format(key))
                            setattr(obj, key, inline_obj)
                            object_updated = True

        if object_updated and commit:
            obj.save()
        return obj


class FormRegister:
    """Class where forms can be registered"""

    def __init__(self):
        """Init"""
        self.forms = defaultdict(dict)

    def add_form(self, code=None, model_alias=None, default_create=False, default_edit=False, minimal=False):
        """Add form to register"""
        def wrapper(form):
            form_code = code if code else form.__name__.lower()
            model_name = self.get_model_alias(model_alias if model_alias else form.Meta.model)

            if form_code in self.forms[model_name]:
                raise Exception("Form {} already registered for model {}".format(code, model_name))

            self.forms[model_name][form_code] = {
                'form': form,
                'default_create': default_create,
                'default_edit': default_edit,
                'minimal': minimal,
            }
            return form

        return wrapper

    def get_model_alias(self, model_alias):
        """Get model alias if class then convert to alias string"""
        from trionyx.models import BaseModel
        if inspect.isclass(model_alias) and issubclass(model_alias, BaseModel):
            config = models_config.get_config(model_alias)
            return '{}.{}'.format(config.app_label, config.model_name)
        return model_alias

    def get_form(self, model, code):
        """Get form based on code"""
        return self.forms.get(self.get_model_alias(model), dict()).get(code)['form']

    def get_all_forms(self, model):
        """Get all forms for model"""
        return [config['form'] for code, config in self.forms.get(self.get_model_alias(model), dict()).items()]

    def get_create_form(self, model):
        """Get default create form"""
        form = self._get_model_form_by_config(model, 'default_create')
        return form if form else self._create_form(model)

    def get_create_minimal_form(self, model):
        """Get default minimal create form"""
        form = self._get_model_form_by_config(model, 'minimal')
        return form if form else self._create_form(model, True)

    def get_edit_form(self, model):
        """Get default edit form"""
        form = self._get_model_form_by_config(model, 'default_edit')
        return form if form else self._create_form(model)

    def _get_model_form_by_config(self, model, config):
        """Get default form"""
        model_alias = self.get_model_alias(model)
        for _, form in self.forms.get(model_alias, dict()).items():
            if form[config]:
                return form['form']
        return None

    def _create_form(self, model, only_required=False):
        """Create form from model"""
        def use_field(field):
            if not only_required:
                return True
            return field.default == NOT_PROVIDED

        config = models_config.get_config(model)
        return modelform_factory(model, fields=[f.name for f in config.get_fields() if use_field(f)])


form_register = FormRegister()
register = form_register.add_form



"""
trionyx.api.filters
~~~~~~~~~~~~~~~~~~~

:copyright: 2019 by Maikel Martens
:license: GPLv3
"""
from django.utils.encoding import force_str
from rest_framework.compat import coreapi, coreschema
from rest_framework.filters import BaseFilterBackend
from rest_framework.settings import api_settings
from watson import search as watson


class SearchFilter(BaseFilterBackend):
    """Search filter that uses watson search"""

    search_param = api_settings.SEARCH_PARAM

    search_title = 'Search'
    search_description = 'A search term.'

    def get_search_term(self, request):
        """
        Search terms are set by a ?search=... query parameter,
        and may be comma and/or whitespace delimited.
        """
        params = request.query_params.get(self.search_param, '')
        return params.replace('\x00', '')  # strip null characters

    def filter_queryset(self, request, queryset, view):
        """Filter queryset"""
        if not self.get_search_term(request):
            return queryset
        return watson.filter(queryset, self.get_search_term(request))

    def get_schema_fields(self, view):
        """Get filter schema fields"""
        assert coreapi is not None, 'coreapi must be installed to use `get_schema_fields()`'
        assert coreschema is not None, 'coreschema must be installed to use `get_schema_fields()`'
        return [
            coreapi.Field(
                name=self.search_param,
                required=False,
                location='query',
                schema=coreschema.String(
                    title=force_str(self.search_title),
                    description=force_str(self.search_description)
                )
            )
        ]

    def get_schema_operation_parameters(self, view):
        """Get schema operation parameters"""
        return [
            {
                'name': self.search_param,
                'required': False,
                'in': 'query',
                'description': force_str(self.search_description),
                'schema': {
                    'type': 'string',
                },
            },
        ]



from trionyx.views import tabs, layouts
from trionyx.layout import Panel, Button, Column12, Table, TableDescription
from trionyx.urls import model_url

from .models import Post, Tag

@layouts.register('view')
@tabs.register(Post)
def post_overview(obj):
    return [
        Column12(
            Panel(
                "Post info",
                TableDescription(
                    'status',
                    'title',
                    'publish_date',
                    'category',
                    'tags',
                    'price',
                    {
                        'label': 'Update',
                        'value': Button(
                            'Delete',
                            dialog_url=model_url(obj, 'dialog-delete'),
                            dialog_options={
                                'callback': "function(){trionyx_reload_tab('general')}"
                            }
                        )
                    }
                )
            )
        ),
        Column12(
            Panel(
                'Tags',
                Table(
                    [
                        ['list object'],
                        {
                            'name': 'dict object',
                        },
                        Tag(name='Tag object'),
                    ],
                    'name=width:120px',
                    'Value=value:fixed value',
                    {
                        'label': 'Value',
                        'value': 'fixed value',
                        'renderer': lambda value, data_object, **options: '{}'.format(data_object)
                    },
                    footer=[
                        [
                            ['Subtotaal', '12,50'],
                            ['Totaal', '12,50'],
                        ],
                        {
                            'colspan': 2,
                            'class': 'text-right'
                        },
                        'Price'
                    ]
                )
            )
        )
    ]

@tabs.register_update('trionyx.user', 'general')
def tasks_update(layout, obj):
    layout.add_component(Panel('Inserted'), path='container.row.column10[1].panels')



"""
trionyx.forms.helper
~~~~~~~~~~~~~~~~~~~~

Helper forms for Trionyx

:copyright: 2018 by Maikel Martens
:license: GPLv3
"""
from crispy_forms.helper import FormHelper as CrispyFormHelper
from trionyx import models
from trionyx import forms
from trionyx.forms import layout
from trionyx.utils import get_datetime_input_format


class FormHelper(CrispyFormHelper):
    """Form helper for Trionyx forms"""

    def build_default_layout(self, form):
        """Build default layout use custom form layouts based on field type"""
        get_model_field = form.Meta.model._meta.get_field if hasattr(form, 'Meta') and hasattr(form.Meta, 'model') else None

        fields = []
        for name, ffield in form.fields.items():
            field = None
            try:
                mfield = get_model_field(name)
            except Exception:
                mfield = None

            # Try get best layout based on model or form field
            if (mfield and isinstance(mfield, models.DateTimeField)) or isinstance(ffield, forms.DateTimeField):
                field = layout.DateTimePicker(name)
            elif (mfield and isinstance(mfield, models.DateField)) or isinstance(ffield, forms.DateField):
                field = layout.DateTimePicker(name, format=get_datetime_input_format(date_only=True))
            elif (mfield and isinstance(mfield, models.TimeField)) or isinstance(ffield, forms.TimeField):
                field = layout.TimePicker(field)

            fields.append(field if field else name)

        return layout.Layout(*fields)



"""
trionyx.renderer
~~~~~~~~~~~~~~~~

:copyright: 2018 by Maikel Martens
:license: GPLv3
"""
import os
from datetime import datetime
from decimal import Decimal
from functools import reduce

from django.conf import settings
from django.core.exceptions import FieldDoesNotExist
from django.utils import formats, timezone
from babel.numbers import format_decimal, format_currency

from trionyx import utils
from trionyx import models


def datetime_value_renderer(value, **options):
    """Render datetime value with django formats, default is SHORT_DATETIME_FORMAT"""
    datetime_format = options.get('datetime_format', 'SHORT_DATETIME_FORMAT')
    if timezone.is_naive(value):
        value = timezone.make_aware(value)
    return formats.date_format(timezone.localtime(value), datetime_format)


def number_value_renderer(value, **options):
    """Format decimal value, with current locale"""
    return format_decimal(value if value else 0.0, locale=utils.get_current_locale())


def price_value_renderer(value, currency=None, **options):
    """Format price value, with current locale and CURRENCY in settings"""
    if not currency:
        currency = getattr(settings, 'CURRENCY', 'USD')
    return format_currency(value if value else 0.0, currency, locale=utils.get_current_locale())


def bool_value_renderer(value, **options):
    """Render boolean value"""
    return '<i class="fa {} {}"></i>'.format(
        'fa-check-square-o' if value else 'fa-square-o',
        'text-success' if value else 'text-danger'
    )


def list_value_renderer(value, **options):
    """Render list value"""
    return ', '.join(value)


def related_field_renderer(value, **options):
    """Render list of related items"""
    return ', '.join(str(obj) for obj in value.all())


def file_field_renderer(file, **options):
    """Render file field as link"""
    if not file:
        return ''

    return '<a href="{file.url}" target="_blank">{name}</a>'.format(file=file, name=os.path.basename(file.path))


def url_field_renderer(value, **options):
    """Render url field"""
    return '<a href="{url}" target="_blank">{url}</a>'.format(url=value) if value else ''


def email_field_renderer(value, **options):
    """Render email field"""
    return '<a href="mailto:{email}" target="_blank">{email}</a>'.format(email=value) if value else ''


class LazyFieldRenderer:
    """Performs render action when __str__ is called"""

    def __init__(self, obj, field_name, **options):
        """Init"""
        self.obj = obj
        self.field_name = field_name
        self.options = options

    def __str__(self):
        """Render field"""
        result = renderer.render_field(self.obj, self.field_name, **self.options)
        return str(result) if result is not None else ''


class Renderer:
    """Registry to hold all renderer's"""

    def __init__(self, renderers=None):
        """Init Renderer"""
        self.renderers = renderers if renderers else {}

    def register(self, type, renderer):
        """Register render function for value or field type"""
        self.renderers[type] = renderer

    def render_value(self, value, **options):
        """Render value"""
        renderer = self.renderers.get(type(value), lambda value, **options: str(value) if value else '')
        return renderer(value, **options)

    def render_field(self, obj, field_name, **options):
        """Render field"""
        field_parts = field_name.split('__')
        if len(field_parts) > 1:
            field_name = field_parts[-1]
            obj = reduce(lambda obj, name: getattr(obj, name), field_parts[:-1], obj)

        try:
            field = obj._meta.get_field(field_name)
        except (FieldDoesNotExist, AttributeError):
            return getattr(obj, field_name, '')

        if hasattr(field, 'choices') and getattr(field, 'choices'):
            return getattr(obj, 'get_{}_display'.format(field_name))()

        value = getattr(obj, field_name, '')
        renderer = self.renderers.get(type(field))

        if renderer:
            return renderer(value, **options)

        return self.render_value(value, **options)


renderer = Renderer({
    datetime: datetime_value_renderer,
    Decimal: number_value_renderer,
    float: number_value_renderer,
    int: number_value_renderer,
    bool: bool_value_renderer,
    list: list_value_renderer,
    models.PriceField: price_value_renderer,
    models.ManyToManyField: related_field_renderer,
    models.FileField: file_field_renderer,
    models.URLField: url_field_renderer,
    models.EmailField: email_field_renderer,
})



"""
trionyx.quickstart
~~~~~~~~~~~~~~~~~~

Quickstart Trionyx project and apps

:copyright: 2017 by Maikel Martens
:license: GPLv3
"""
import os
import shutil

import trionyx
from trionyx import utils


class Quickstart:
    """Class for creating Trionyx project and apps"""

    def __init__(self):
        """Set project template paths"""
        quickstart_path = os.path.realpath(__file__)

        self.project_path = os.path.join(os.path.dirname(quickstart_path), 'project')
        """Path to project template files"""

        self.app_path = os.path.join(os.path.dirname(quickstart_path), 'app')
        """Path to app template files"""

    def create_project(self, project_path):
        """
        Create Trionyx project in given path

        :param str path: path to create project in.
        :raises FileExistsError:
        """
        shutil.copytree(self.project_path, project_path)

        self.update_file(project_path, 'requirements.txt', {
            'trionyx_version': trionyx.__version__
        })

        self.update_file(project_path, 'environment.json', {
            'secret_key': utils.random_string(32)
        })

    def create_app(self, apps_path, name):
        """
        Create Trionyx app in given path

        :param str path: path to create app in.
        :param str name: name of app
        :raises FileExistsError:
        """
        app_path = os.path.join(apps_path, name.lower())

        shutil.copytree(self.app_path, app_path)

        self.update_file(app_path, '__init__.py', {
            'name': name.lower()
        })

        self.update_file(app_path, 'apps.py', {
            'name': name.lower(),
            'verbose_name': name.capitalize()
        })

    def update_file(self, project_path, file_path, variables):
        """
        Update given file with given variables, variables in file must be inclosed with [[]].

        For example you want to replace a variable secret_key, in the file its [[secret_key]].

        :param str project_path:
        :param str file_path:
        :param dict variables:
        """
        update_file = os.path.join(project_path, file_path)
        with open(update_file, 'rb') as _file:
            file_content = _file.read().decode('utf-8')

        for key, value in variables.items():
            file_content = file_content.replace('[[{}]]'.format(key), value)

        with open(update_file, 'w+', encoding='utf-8') as _file:
            _file.writelines(file_content)



# Generated by Django 2.2.6 on 2019-10-29 19:50

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('testblog', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='tag',
            name='created_by',
            field=models.ForeignKey(blank=True, default=None, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to=settings.AUTH_USER_MODEL, verbose_name='Created by'),
        ),
        migrations.AddField(
            model_name='post',
            name='category',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='posts', to='testblog.Category'),
        ),
        migrations.AddField(
            model_name='post',
            name='created_by',
            field=models.ForeignKey(blank=True, default=None, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to=settings.AUTH_USER_MODEL, verbose_name='Created by'),
        ),
        migrations.AddField(
            model_name='post',
            name='tags',
            field=models.ManyToManyField(related_name='posts', to='testblog.Tag'),
        ),
        migrations.AddField(
            model_name='category',
            name='created_by',
            field=models.ForeignKey(blank=True, default=None, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to=settings.AUTH_USER_MODEL, verbose_name='Created by'),
        ),
    ]



"""
trionyx.urls
~~~~~~~~~~~~

:copyright: 2017 by Maikel Martens
:license: GPLv3
"""
import importlib
import inspect
from pkg_resources import iter_entry_points
from django.conf import settings
from django.apps import apps
from django.conf.urls import include
from django.conf.urls.static import static
from django.urls import path, reverse, NoReverseMatch


def model_url(model, view_name, code=None):
    """Shortcut function for getting model url"""
    from trionyx.config import models_config
    view_name = 'trionyx:model-{}'.format(view_name)
    config = models_config.get_config(model)
    kwargs = {
        'app': config.app_label,
        'model': config.model_name,
    }
    if code:
        kwargs['code'] = code

    if not inspect.isclass(model) and not isinstance(model, str):
        try:
            kwargs['pk'] = model.pk
            return reverse(view_name, kwargs=kwargs)
        except NoReverseMatch:
            kwargs.pop('pk')
    return reverse(view_name, kwargs=kwargs)


urlpatterns = [
    path('', include('trionyx.trionyx.urls', namespace='trionyx')),
]

# Add installed apps urls
for entry_point in iter_entry_points(group='trionyx.app', name=None):
    if importlib.util.find_spec(entry_point.module_name + '.urls'):
        urlpatterns.append(
            path('', include(entry_point.module_name + '.urls'))
        )

if settings.DEBUG:
    urlpatterns += static(
        settings.STATIC_URL, document_root=settings.STATIC_ROOT
    )

if settings.DEBUG and apps.is_installed("debug_toolbar"):
    import debug_toolbar
    urlpatterns += [
        path('__debug__/', include(debug_toolbar.urls)),
    ]



"""
trionyx.trionyx.api.serializers
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

:copyright: 2019 by Maikel Martens
:license: GPLv3
"""
import inspect

from trionyx.config import models_config


class SerializerRegister:
    """SerializerRegister class"""

    def __init__(self):
        """Init class"""
        self.serializers = {}

    def register(self, model_alias):
        """Add form to register"""
        def wrapper(serializer):
            model_name = self.get_model_alias(model_alias)

            if model_name in self.serializers:
                raise Exception("Serializer {} already registered for model {}".format(serializer, model_name))

            self.serializers[model_name] = serializer
            return serializer

        return wrapper

    def get_model_alias(self, model_alias):
        """Get model alias if class then convert to alias string"""
        from trionyx.models import BaseModel
        if inspect.isclass(model_alias) and issubclass(model_alias, BaseModel):
            config = models_config.get_config(model_alias)
            return '{}.{}'.format(config.app_label, config.model_name)
        return model_alias

    def get(self, model_alias):
        """Get serializer based on model or model alias"""
        return self.serializers.get(self.get_model_alias(model_alias))


serializer_register = SerializerRegister()
register = serializer_register.register



"""
trionyx.trionyx.middleware
~~~~~~~~~~~~~~~~~~~~~~~~~~

:copyright: 2018 by Maikel Martens
:license: GPLv3
"""
import threading
from re import compile

from django.urls import reverse
from django.http import HttpResponseRedirect
from django.conf import settings
from django.utils import timezone
from django.utils import translation

LOCAL_DATA = threading.local()

EXEMPT_URLS = [
    compile(reverse(settings.LOGIN_URL).lstrip('/')),
]
if hasattr(settings, 'LOGIN_EXEMPT_URLS'):
    EXEMPT_URLS += [compile(expr) for expr in settings.LOGIN_EXEMPT_URLS]


class LoginRequiredMiddleware:
    """
    Middleware that requires a user to be authenticated to view any page other
    than LOGIN_URL. Exemptions to this requirement can optionally be specified
    in settings via a list of regular expressions in LOGIN_EXEMPT_URLS (which
    you can copy from your urls.py).

    Requires authentication middleware and template context processors to be
    loaded. You'll get an error if they aren't.
    """

    def __init__(self, get_response):
        """Init"""
        self.get_response = get_response

    def __call__(self, request):
        """Check if user is logged in"""
        if not request.user.is_authenticated:
            path = request.path_info.lstrip('/')
            if not any(m.match(path) for m in EXEMPT_URLS):
                return HttpResponseRedirect(reverse(settings.LOGIN_URL))

        return self.get_response(request)


class GlobalRequestMiddleware:
    """Store request in thread local data"""

    def __init__(self, get_response):
        """Init"""
        self.get_response = get_response

    def __call__(self, request):
        """Store request in local data"""
        LOCAL_DATA.request = request
        try:
            return self.get_response(request)
        finally:
            del LOCAL_DATA.request


class LastLoginMiddleware:
    """Set last login for user"""

    def __init__(self, get_response):
        """Init"""
        self.get_response = get_response

    def __call__(self, request):
        """Set last login"""
        minute_ago = timezone.now() - timezone.timedelta(minutes=1)
        if request.user.is_authenticated and (not request.user.last_online or request.user.last_online <= minute_ago):
            request.user.last_online = timezone.now()
            request.user.save(update_fields=['last_online'])

        return self.get_response(request)


class LocalizationMiddleware:
    """Localize request to user settings"""

    def __init__(self, get_response):
        """Init"""
        self.get_response = get_response

    def __call__(self, request):
        """Set localization"""
        if request.user.is_authenticated:
            translation.activate(request.user.language)
            timezone.activate(request.user.timezone)

        return self.get_response(request)



"""
trionyx.trionyx.view.accounts
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

:copyright: 2018 by Maikel Martens
:license: GPLv3
"""
import json
import logging
from collections import OrderedDict

from django.http import HttpResponse, HttpResponseRedirect
from django.contrib.auth.views import LoginView as DjangoLoginView
from django.contrib.auth import logout as django_logout
from django.contrib.auth.models import Permission
from django.shortcuts import redirect
from django.urls import reverse
from django_jsend import JsendView
from watson import search as watson
from django.contrib.contenttypes.models import ContentType
from django.views.generic import TemplateView
from django.conf import settings
from django.contrib.staticfiles.templatetags.staticfiles import static
from django.utils.translation import ugettext_lazy as _
from django.contrib import messages

from trionyx.models import get_class
from trionyx.views import UpdateView, DetailTabView, DialogView
from trionyx.views.mixins import ModelClassMixin, ModelPermissionMixin
from trionyx.config import models_config
from trionyx.widgets import widgets
from trionyx import utils
from trionyx.forms.helper import FormHelper
from trionyx.forms import form_register, modelform_factory
from trionyx.models import filter_queryset_with_user_filters
from trionyx.trionyx.tasks import MassUpdateTask

User = get_class('trionyx.User')
Task = get_class('trionyx.Task')

logger = logging.getLogger(__name__)


def media_nginx_accel(request, path):
    """
    Location /protected/ {
        internal;
        root <complete path to project root dir>;
    }

    """
    response = HttpResponse(status=200)
    response['Content-Type'] = ''
    response['X-Accel-Redirect'] = '/protected' + request.path
    return response


class LoginView(DjangoLoginView):
    """Trionyx login view"""

    template_name = 'trionyx/core/login.html'


def logout(request):
    """Trionyx logout view"""
    django_logout(request)
    return redirect('/')


class UpdateUserAccountView(UpdateView):
    """Update user view"""

    model = User
    title = 'Update account'
    cancel_url = 'trionyx:view-account'

    @property
    def form_class(self):
        """Form class used by view"""
        from trionyx.trionyx.forms import ProfileUpdateForm
        return ProfileUpdateForm

    def get(self, request, *args, **kwargs):
        """Add user id to kwargs"""
        kwargs['pk'] = request.user.id
        self.kwargs['pk'] = request.user.id
        return super().get(request, *args, **kwargs)

    def post(self, request, *args, **kwargs):
        """Add user id to kwargs"""
        kwargs['pk'] = request.user.id
        self.kwargs['pk'] = request.user.id
        return super().post(request, *args, **kwargs)

    def get_success_url(self):
        """Set success url to account page"""
        return reverse('trionyx:view-account')


class ViewUserAccountView(DetailTabView):
    """View user"""

    model_alias = 'trionyx.profile'
    model = User

    title = 'Profile'

    def get(self, request, *args, **kwargs):
        """Add user id to kwargs"""
        kwargs['pk'] = request.user.id
        self.kwargs['pk'] = request.user.id
        return super().get(request, *args, **kwargs)

    def get_edit_url(self):
        """Get user edit url"""
        return reverse('trionyx:edit-account')

    def get_delete_url(self):
        """Hide delete url"""
        return None

    def get_back_url(self):
        """Hide back url"""
        return None


def create_permission_jstree(selected=None, disabled=False):
    """Create permission jstree"""
    selected = selected if selected else []
    jstree = []

    added_apps = ['auth']
    added_models = []
    for permission in Permission.objects.select_related('content_type').all():
        if not permission.content_type.model_class():
            continue

        model_config = models_config.get_config(permission.content_type.model_class())

        if model_config.hide_permissions:
            continue

        if model_config.disable_add and permission.codename == 'add_{}'.format(model_config.model_name):
            continue

        if model_config.disable_change and permission.codename == 'change_{}'.format(model_config.model_name):
            continue

        if model_config.disable_delete and permission.codename == 'delete_{}'.format(model_config.model_name):
            continue

        if model_config.app_label not in added_apps:
            jstree.append({
                'id': model_config.app_label,
                'parent': '#',
                'text': model_config.get_app_verbose_name(),
                'state': {
                    'disabled': disabled,
                }
            })
            added_apps.append(model_config.app_label)

        parent = [model_config.app_label if model_config.app_label != 'auth' else 'trionyx']

        if model_config.model_name not in added_models:
            jstree.append({
                'id': '.'.join([*parent, model_config.model_name]),
                'parent': '.'.join(parent),
                'text': model_config.get_verbose_name_plural(),
                'state': {
                    'disabled': disabled,
                }
            })
            added_models.append(model_config.model_name)

        parent.append(model_config.model_name)
        name = {
            'view_{}'.format(model_config.model_name): _('View'),
            'add_{}'.format(model_config.model_name): _('Add'),
            'change_{}'.format(model_config.model_name): _('Change'),
            'delete_{}'.format(model_config.model_name): _('Delete'),
        }.get(permission.codename, permission.name)

        jstree.append({
            'id': '.'.join([*parent, permission.codename]),
            'parent': '.'.join(parent),
            'text': str(name),
            'state': {
                'selected': permission in selected,
                'disabled': disabled,
            },
            'permission_id': permission.id,
        })

    return jstree


# =============================================================================
# Global search
# =============================================================================
class GlobalSearchJsendView(JsendView):
    """View for global search uses watson to search all models"""

    def handle_request(self, request):
        """Handle search"""
        models = []
        content_types = {}
        for config in models_config.get_all_configs(False):
            if config.disable_search_index or not config.global_search:
                continue

            # Check if user has view permission
            if not request.user.has_perm('{app_label}.view_{model_name}'.format(
                app_label=config.app_label,
                model_name=config.model_name,
            ).lower()):
                continue

            models.append(config.model)
            content_type = ContentType.objects.get_for_model(config.model, False)
            content_types[content_type.id] = str(config.model._meta.verbose_name_plural)

        results = OrderedDict()
        for entry in watson.search(request.GET.get('search', ''), models=models)[:100]:
            if entry.content_type_id not in results:
                results[entry.content_type_id] = {
                    'name': content_types[entry.content_type_id],
                    'items': [],
                }

            if len(results[entry.content_type_id]['items']) <= 10:
                results[entry.content_type_id]['items'].append({
                    'url': entry.url,
                    'title': entry.title,
                    'description': entry.description,
                })

        return list(results.values())


class FilterFieldsJsendView(JsendView):
    """View for getting model filter fields"""

    def handle_request(self, request, *args, **kwargs):
        """Get filter fields"""
        modelClass = ContentType.objects.get_for_id(request.GET.get('id')).model_class()
        config = models_config.get_config(modelClass)

        return {
            'id': request.GET.get('id'),
            'fields': {
                name: {
                    'name': name,
                    'label': str(field['label']),
                    'type': field['type'],
                    'choices': field['choices'],
                    'choices_url': field.get('choices_url', None)
                }
                for name, field in config.get_list_fields().items()
            }
        }


class UserTasksJsend(JsendView):
    """User tasks view"""

    def handle_request(self, request):
        """Get user open tasks"""
        tasks = []

        tasks.extend([
            task for task in Task.objects.filter(user=request.user, status=Task.SCHEDULED).order_by('-scheduled_at')
        ])

        tasks.extend([
            task for task in Task.objects.filter(
                user=request.user,
                status__in=[Task.QUEUE, Task.LOCKED, Task.RUNNING]
            ).order_by('-started_at')
        ])

        if len(tasks) < 10:
            tasks.extend([
                task for task in Task.objects.filter(
                    user=request.user, status__in=[Task.COMPLETE, Task.FAILED]
                ).order_by('-started_at')[: 10 - len(tasks)]
            ])

        return [
            {
                'id': task.id,
                'status': task.status,
                'status_display': task.get_status_display(),
                'description': task.description,
                'progress': task.progress,
                'url': task.get_absolute_url(),
            } for task in tasks
        ]


# =============================================================================
# Dashboard
# =============================================================================
class DashboardView(TemplateView):
    """Dashboard view"""

    template_name = 'trionyx/core/dashboard.html'

    def get_context_data(self, *args, **kwargs):
        """Get dashboard context data"""
        context = super().get_context_data(*args, ** kwargs)
        dashboard = self.request.user.get_attribute('tx_dashboard', [])
        if not dashboard:
            for widget in (
                settings.TX_DEFAULT_DASHBOARD()
                if callable(settings.TX_DEFAULT_DASHBOARD)
                else settings.TX_DEFAULT_DASHBOARD
            ):
                widget = widget.copy()
                widget['i'] = utils.random_string(16)
                dashboard.append(widget)

            self.request.user.set_attribute('tx_dashboard', dashboard)

        context.update({
            'widget_templates': [widget().template for index, widget in widgets.items()],
            'widgets': [{
                'code': widget.code,
                'name': str(widget.name),
                'description': str(widget.description),
                'image': static(widget().image),
                'config_fields': widget().config_fields,
                'default_w': widget.default_width,
                'default_h': widget.default_height,
            } for index, widget in widgets.items()],
            'dashboard': dashboard,
        })
        return context


class WidgetDataJsendView(JsendView):
    """Jsend view to get widget data"""

    def handle_request(self, request):
        """Get widget data"""
        data = json.loads(request.body.decode('utf-8'))

        if data['code'] not in widgets:
            raise Exception('Widget does not exists')

        widget = widgets.get(data['code'])()

        return widget.get_data(request, data.get('config', {}))


class SaveDashboardJsendView(JsendView):
    """Jsend view to save user dashboard"""

    def handle_request(self, request):
        """Handle dashboard save"""
        dashboard = json.loads(request.body.decode('utf-8'))

        if not isinstance(dashboard, list):
            raise Exception('Expect dashboard to be a list')

        request.user.set_attribute('tx_dashboard', dashboard)

        return 'saved'


class WidgetConfigDialog(DialogView):
    """Widget config dialog view"""

    def display_dialog(self, code, *args, **kwargs):
        """Handle widget config"""
        if code not in widgets:
            raise Exception('Widget does not exists')

        if '__post__' in self.request.POST:
            post = self.request.POST
            initial = None
        else:
            post = None
            initial = self.request.POST

        widget = widgets.get(code)
        form = widget.config_form_class(post, initial=initial) if widget.config_form_class else None

        config = None
        if form:
            if not hasattr(form, "helper"):
                form.helper = FormHelper(form)
            form.helper.form_tag = False

            if form.is_valid():
                config = form.cleaned_data
                config['title'] = self.request.POST.get('title')
                config['refresh'] = self.request.POST.get('refresh')
            else:
                logger.error(json.dumps(form.errors))
        elif '__post__' in self.request.POST:
            config = {
                'title': self.request.POST.get('title'),
                'refresh': self.request.POST.get('refresh')
            }

        return {
            'title': _('Widget config'),
            'content': self.render_to_string('trionyx/dialog/widget_config.html', {
                'form': form,
                'title': self.request.POST.get('title'),
                'refresh': self.request.POST.get('refresh'),
            }),
            'submit_label': _('Save'),
            'config': config
        }

    def handle_dialog(self, code, *args, **kwargs):
        """Handle widget config"""
        return self.display_dialog(code)


# Mass actions
class MassDeleteDialog(DialogView):
    """Mass delete dialog"""

    permission_type = 'delete'

    def has_permission(self, request):
        """Check permission"""
        config = self.get_model_config()

        if config.disable_delete:
            return False

        return request.user.has_perm('{app_label}.{type}_{model_name}'.format(
            app_label=config.app_label,
            type=self.permission_type,
            model_name=config.model_name,
        ).lower())

    def display_dialog(self, *args, **kwargs):
        """Handle mass delete"""
        query = self.get_model_class().objects.get_queryset()

        if self.request.POST.get('all', '0') == '1':
            query = filter_queryset_with_user_filters(query, json.loads(self.request.POST.get('filters', '[]')))
        else:
            query = query.filter(id__in=[int(id) for id in filter(None, self.request.POST.get('ids', '').split(','))])

        count = query.count()
        if '__post__' in self.request.POST:
            try:
                query.delete()
            except Exception:
                return {
                    'title': _('Something went wrong on deleting the items'),
                }
            return {
                'deleted': True,
            }
        else:
            return {
                'title': _('Deleting {count} {model_name} items').format(
                    count=count,
                    model_name=self.get_model_config().get_verbose_name(False)
                ),
                'content': self.render_to_string('trionyx/dialog/mass_delete.html', {
                    'count': count,
                    'model_name': self.get_model_config().get_verbose_name(False),
                    'all': self.request.POST.get('all', '0'),
                    'filters': self.request.POST.get('filters', '[]'),
                    'ids': self.request.POST.get('ids', ''),
                }),
                'submit_label': _('Delete') if count else None,
            }

    def handle_dialog(self, *args, **kwargs):
        """Handle mass delete"""
        return self.display_dialog()


class MassUpdateView(ModelPermissionMixin, TemplateView, ModelClassMixin):
    """Mass update view"""

    template_name = 'trionyx/core/model_mass_update.html'
    permission_type = 'change'

    def get(self, *args, **kwargs):
        """Render mass update form"""
        context = self.get_context_data(**kwargs)
        all = self.request.GET.get('all', '0')
        ids = self.request.GET.get('ids', '')
        filters = self.request.GET.get('filters', '[]')
        query = self.get_queryset(all, ids, filters)

        if not query:
            messages.error(self.request, _('You must make a selection'))
            return HttpResponseRedirect(reverse('trionyx:model-list', kwargs=self.kwargs))

        context.update({
            'model_name': self.get_model_config().get_verbose_name(),
            'model_name_plural': self.get_model_config().get_verbose_name_plural(),
            'cancel_url': reverse('trionyx:model-list', kwargs=self.kwargs),
            'all': all,
            'ids': ids,
            'filters': filters,
            'count': query.count(),
            'form': self.get_form(),
            'checked_fields': [],
        })
        return self.render_to_response(context)

    def post(self, *args, **kwargs):
        """Handle mass update form"""
        all = self.request.POST.get('trionyx_all', '0')
        ids = self.request.POST.get('trionyx_ids', '')
        filters = self.request.POST.get('trionyx_filters', '[]')
        query = self.get_queryset(all, ids, filters)

        form = self.get_form(self.request.POST)

        if not form.is_valid():
            context = self.get_context_data(**kwargs)
            context.update({
                'model_name': self.get_model_config().get_verbose_name(),
                'model_name_plural': self.get_model_config().get_verbose_name_plural(),
                'cancel_url': reverse('trionyx:model-list', kwargs=self.kwargs),
                'all': all,
                'ids': ids,
                'filters': filters,
                'count': query.count(),
                'form': form,
                'checked_fields': form.checked_fields,
            })
            return self.render_to_response(context)

        # Get data from form
        data = {}
        for field in form.fields:
            if self.request.POST.get('change_{}'.format(field), False):
                data[field] = form.cleaned_data[field]

        # Start update task
        MassUpdateTask().delay(
            all=all,
            ids=ids,
            filters=filters,
            data=data,

            task_description=_('Mass update {count} {model_name}').format(
                count=query.count(),
                model_name=self.get_model_config().get_verbose_name_plural(),
            ),
            task_model=self.get_model_class(),
        )

        # Add message
        messages.success(self.request, _('Successfully started task for updating {count} {model_name}').format(
            count=query.count(),
            model_name=self.get_model_config().get_verbose_name_plural(),
        ))

        # return to list view
        return HttpResponseRedirect(reverse('trionyx:model-list', kwargs=self.kwargs))

    def get_queryset(self, all, ids, filters):
        """Get queryset"""
        query = self.get_model_class().objects.get_queryset()

        if all == '1':
            return filter_queryset_with_user_filters(query, json.loads(filters))
        else:
            return query.filter(id__in=[int(id) for id in filter(None, ids.split(','))])

    def get_form(self, data=None):
        """Get form class"""
        model_config = self.get_model_config()
        fields = []

        model_fields = [f.name for f in model_config.get_fields()]
        forms = form_register.get_all_forms(self.get_model_class())
        if forms:  # Only use fields that ara also in forms
            for Form in forms:
                fields.extend([name for name in Form().base_fields if name in model_fields])
                fields.extend([name for name in Form().declared_fields if name in model_fields])
        else:
            fields = [f.name for f in model_config.get_fields()]

        FormClass = modelform_factory(self.get_model_class(), fields=sorted(list(set(fields))))
        form = FormClass(data)

        # Disable fields that are not checked for change
        form.checked_fields = []
        for field in form.fields:
            form.fields[field].required = False
            if not (data and data.get('change_{}'.format(field), False)):
                form.fields[field].widget.attrs['readonly'] = True
                form.fields[field].widget.attrs['disabled'] = True
            else:
                form.checked_fields.append(field)

        return form



"""
trionyx.views.dialogs
~~~~~~~~~~~~~~~~~~~~~

:copyright: 2018 by Maikel Martens
:license: GPLv3
"""
import logging
import json

from django.views.generic import View
from django.http import JsonResponse
from django.template.loader import render_to_string
from django.utils.translation import ugettext_lazy as _

from trionyx.views.mixins import ModelClassMixin
from trionyx.forms.helper import FormHelper

logger = logging.getLogger(__name__)


class DialogView(View, ModelClassMixin):
    """
    Dialog view used for showing dialog popup with ajax.

    **Example**

    .. code-block:: python

        # apps/your_app/views.py

        class HelloWorldDialogView(DialogView):

            def display_dialog(self):
                return {
                    "title": "Hello world",
                    "content": self.render_to_string('your_app/dialog/hello_world.html'),
                }

    **Calling in frontend**

    .. code-block:: django

        <button class="btn btn-primary btn-block"
            onClick="openDialog('{% url 'hello:world' %}', {/* dialog options */})">
            Hello World
        </button>
    """

    permission_type = 'view'
    """Permission type used for model when no permission is set"""

    permission = None
    """Permission of dialog, when not set and model is set model view permission is used"""

    model_permission = 'read'
    """When no permission is set will use this for model permission"""

    model = None
    """
    Model that will be auto loaded, when set the url mus contain a `pk` parameter.
    Model can also be set with the url with the `app` and `model` parameter.

    Loaded model object is available as self.object and self.<object name lowercase>
    """

    def get(self, request, *args, **kwargs):
        """Handle get request"""
        try:
            kwargs = self.load_object(kwargs)
        except Exception as e:
            return self.render_te_response({
                'title': str(e),
            })

        if not self.has_permission(request):
            return self.render_te_response({
                'title': _('No access'),
            })
        return self.render_te_response(self.display_dialog(*args, **kwargs))

    def post(self, request, *args, **kwargs):
        """Handle post request"""
        try:
            kwargs = self.load_object(kwargs)
        except Exception as e:
            return self.render_te_response({
                'title': str(e),
            })

        if not self.has_permission(request):
            return self.render_te_response({
                'title': _('No access'),
            })
        return self.render_te_response(self.handle_dialog(*args, **kwargs))

    def load_object(self, kwargs):
        """Load object and model config and remove pk from kwargs"""
        self.object = None
        self.config = None
        self.model = self.get_model_class()
        kwargs.pop('app', None)
        kwargs.pop('model', None)

        if self.model and kwargs.get('pk', False):
            try:
                self.object = self.model.objects.get(pk=kwargs.pop('pk'))
            except Exception:
                raise Exception(_("Could not load {model}").format(model=self.model.__name__.lower()))
            setattr(self, self.model.__name__.lower(), self.object)

        return kwargs

    def has_permission(self, request):
        """Check if user has permission"""
        if not self.object and not self.permission:
            return True

        if not self.permission:
            config = self.get_model_config()

            if self.permission_type == 'change' and config.disable_change:
                return False

            if self.permission_type == 'add' and config.disable_add:
                return False

            if self.permission_type == 'delete' and config.disable_delete:
                return False

            return request.user.has_perm('{app_label}.{type}_{model_name}'.format(
                app_label=config.app_label,
                type=self.permission_type,
                model_name=config.model_name,
            ).lower())

        return request.user.has_perm(self.permission)

    def render_to_string(self, template_file, context):
        """Render given template to string and add object to context"""
        context = context if context else {}
        if self.object:
            context['object'] = self.object
            context[self.object.__class__.__name__.lower()] = self.object
        return render_to_string(template_file, context, self.request)

    def display_dialog(self, *args, **kwargs):
        """
        Override this function to display a dialog popup. Url params are given as function params.

        The function must return a dict that can contain the following data:

        - **title**: Title of the dialog
        - **content**: Html content to display in dialog, shortcut function :class:`trionyx.trionyx.views.core.DialogView.render_to_string`
        - **url (optional)**: Post url for dialog form, must be a link to a DialogView.
        - **submit_label (optional)**: Label of the submit button, when empty submit button is not shown.
        - **redirect_url (optional)**: Redirect page to given url.
        - **close (optional)**: Close dialog.
        """
        return {}

    def handle_dialog(self, *args, **kwargs):
        """
        Override this function to handle and display popup. Url params are given as function params.
        This function must return the same dict structure as :class:`trionyx.trionyx.views.core.DialogView.display_dialog`

        Post data can be retrieved with *self.request.POST*
        """
        return {}

    def render_te_response(self, data):
        """Render data to JsonResponse"""
        if 'submit_label' in data and 'url' not in data:
            data['url'] = self.request.get_full_path()

        return JsonResponse(data)


class LayoutDialog(DialogView):
    """Render layout in dialog"""

    def display_dialog(self, code):
        """Render layout for object"""
        from trionyx.views import layouts
        try:
            content = layouts.get_layout(code, self.object, self.request)
        except Exception:
            content = _('Layout does not exists')

        return {
            'title': str(self.object),
            'content': content,
        }


class UpdateDialog(DialogView):
    """Update dialog view for updating a model"""

    permission_type = 'change'

    template = 'trionyx/dialog/model_form.html'
    """Template for dialog content"""

    title = _("Update {model_name}: {object}")
    """Dialog title model_name and object, variable are given"""

    submit_label = _('save')
    """Dialog submit label value"""

    success_message = _('{model_name} ({object}) is successfully updated')
    """Success message on successfully form saved"""

    def get_form_class(self):
        """Get form class for dialog, default will get form from model config"""
        from trionyx.forms import form_register
        if self.kwargs.get('code'):
            return form_register.get_form(self.get_model_class(), self.kwargs.get('code'))
        return form_register.get_edit_form(self.get_model_class())

    def display_dialog(self, *args, **kwargs):
        """Display form and success message when set"""
        form = kwargs.pop('form_instance', None)
        success_message = kwargs.pop('success_message', None)

        if not form:
            form = self.get_form_class()(initial=kwargs, instance=self.object)

        if not hasattr(form, "helper"):
            form.helper = FormHelper(form)
        form.helper.form_tag = False

        return {
            'title': self.title.format(
                model_name=self.get_model_config().model_name,
                object=str(self.object) if self.object else '',
            ),
            'content': self.render_to_string(self.template, {
                'form': form,
                'success_message': success_message,
            }),
            'submit_label': self.submit_label,
            'success': bool(success_message),
        }

    def handle_dialog(self, *args, **kwargs):
        """Handle form and save and set success message on valid form"""
        form = self.get_form_class()(self.request.POST, initial=kwargs, instance=self.object)

        success_message = None
        if form.is_valid():
            obj = form.save()
            success_message = self.success_message.format(
                model_name=self.get_model_config().model_name.capitalize(),
                object=str(obj),
            )
        else:
            logger.debug(json.dumps(form.errors))

        return self.display_dialog(*args, form_instance=form, success_message=success_message, **kwargs)


class CreateDialog(UpdateDialog):
    """Dialog view for creating a model"""

    permission_type = 'add'

    title = _("Create {model_name}")
    submit_label = _('create')
    success_message = _('{model_name} ({object}) is successfully created')

    def get_form_class(self):
        """Get create form class"""
        from trionyx.forms import form_register
        if self.kwargs.get('code'):
            return form_register.get_form(self.get_model_class(), self.kwargs.get('code'))
        return form_register.get_create_form(self.get_model_class())


class DeleteDialog(DialogView):
    """Delete model dialog"""

    permission_type = 'delete'

    template = 'trionyx/dialog/model_delete.html'
    """Template for dialog content"""

    title = _("Delete {model_name}: {object}")
    """Dialog title model_name and object, variable are given"""

    submit_label = _('delete')
    """Dialog submit label value"""

    success_message = _('{model_name} ({object}) is successfully deleted')
    """Success message on successfully deleted"""

    def display_dialog(self, *args, **kwargs):
        """Display delete confirmation"""
        return {
            'title': self.title.format(
                model_name=self.get_model_config().model_name,
                object=str(self.object),
            ),
            'content': self.render_to_string(self.template, {
                'model_name': self.get_model_config().model_name.capitalize(),
                'object': self.object,
            }),
            'submit_label': self.submit_label,
        }

    def handle_dialog(self, *args, **kwargs):
        """Handle delete model"""
        object_name = str(self.object)
        response = {
            'title': self.title.format(
                model_name=self.get_model_config().model_name,
                object=object_name,
            ),
        }

        try:
            self.object.delete()
            response.update({
                'content': self.success_message.format(
                    model_name=self.get_model_config().model_name.capitalize(),
                    object=object_name,
                )
            })
        except Exception:
            response.update({
                'content': """<p class="alert alert-danger">{}</p>""".format(
                    _('Something went wrong on deleting {model_name}').format(
                        model_name=self.get_model_config().model_name.capitalize())
                )
            })

        return response



# Generated by Django 2.2.6 on 2019-10-29 19:50

from django.db import migrations, models
import trionyx.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Category',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, verbose_name='Created at')),
                ('updated_at', models.DateTimeField(auto_now=True, verbose_name='Updated at')),
                ('deleted', models.BooleanField(default=False, verbose_name='Deleted')),
                ('verbose_name', models.TextField(blank=True, default='', verbose_name='Verbose name')),
                ('name', models.CharField(max_length=255)),
                ('description', models.TextField(default='')),
            ],
            options={
                'verbose_name_plural': 'Categories',
            },
        ),
        migrations.CreateModel(
            name='Post',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, verbose_name='Created at')),
                ('updated_at', models.DateTimeField(auto_now=True, verbose_name='Updated at')),
                ('deleted', models.BooleanField(default=False, verbose_name='Deleted')),
                ('verbose_name', models.TextField(blank=True, default='', verbose_name='Verbose name')),
                ('title', models.CharField(max_length=255)),
                ('content', models.TextField()),
                ('publish_date', models.DateTimeField()),
                ('status', models.IntegerField(choices=[(1, 'New'), (2, 'Draft'), (3, 'Published')], default=1)),
                ('price', trionyx.models.PriceField(decimal_places=4, default=0.0, max_digits=11)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Tag',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, verbose_name='Created at')),
                ('updated_at', models.DateTimeField(auto_now=True, verbose_name='Updated at')),
                ('deleted', models.BooleanField(default=False, verbose_name='Deleted')),
                ('verbose_name', models.TextField(blank=True, default='', verbose_name='Verbose name')),
                ('name', models.CharField(max_length=255)),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



"""
trionyx.trionyx.auditlog
~~~~~~~~~~~~~~~~~~~~~~~~

:copyright: 2019 by Maikel Martens
:license: GPLv3
"""
import logging

from django.conf import settings
from django.utils import timezone
from django.core.exceptions import ObjectDoesNotExist
from django.db.models.signals import pre_save, post_save, post_delete
from django.utils.translation import ugettext_lazy as _
from trionyx import models
from trionyx.config import models_config
from trionyx.trionyx.models import AuditLogEntry
from trionyx.utils import get_current_request
from trionyx.trionyx.layouts import auditlog as auditlog_layout
from trionyx.views import tabs
from trionyx.renderer import renderer

logger = logging.getLogger(__name__)


def model_instance_diff(old, new):
    """Create diff of two model instances"""
    diff = {}

    config = models_config.get_config(new if new else old)
    ignore_fields = config.auditlog_ignore_fields if config.auditlog_ignore_fields else []
    fields = [field for field in config.get_fields() if field.name not in ignore_fields]

    for field in fields:
        old_value = get_field_value(old, field)
        new_value = get_field_value(new, field)

        if old_value != new_value:
            diff[field.name] = (
                get_rendered_value(config.model, field.name, old_value),
                get_rendered_value(config.model, field.name, new_value),
            )

    return diff if diff else None


def get_field_value(obj, field):
    """Get field value that can be used to compare"""
    if isinstance(field, models.DateTimeField):
        try:
            value = field.to_python(getattr(obj, field.name, None))
            if value is not None and settings.USE_TZ and not timezone.is_naive(value):
                value = timezone.make_naive(value, timezone=timezone.utc)
            return value
        except ObjectDoesNotExist:
            pass
    elif isinstance(field, models.DecimalField):
        try:
            return field.to_python(getattr(obj, field.name, None))
        except TypeError:
            pass
    else:
        try:
            return getattr(obj, field.name, None)
        except ObjectDoesNotExist:
            pass

    return field.default if field.default is not models.NOT_PROVIDED else None


def get_rendered_value(ModelClass, field_name, value):
    """Render value for given model class and field"""
    model = ModelClass(**{field_name: value})
    return renderer.render_field(model, field_name)


def create_log(instance, changes, action):
    """Create a new log entry"""
    AuditLogEntry.objects.create(
        content_object=instance,
        object_verbose_name=str(instance),
        action=action,
        changes=changes,
        user=get_current_request().user if get_current_request() and not get_current_request().user.is_anonymous else None
    )


def log_add(sender, instance, created, **kwargs):
    """Log model add"""
    try:
        if created:
            changes = model_instance_diff(None, instance)
            if changes:
                create_log(instance, changes, AuditLogEntry.ACTION_ADDED)
    except Exception as e:
        logger.exception(e)


def log_change(sender, instance, **kwargs):
    """Log model update"""
    try:
        if instance.pk is not None:
            try:
                old = sender.objects.get(pk=instance.pk)
            except sender.DoesNotExist:
                return

            changes = model_instance_diff(old, instance)
            if changes:
                create_log(instance, changes, AuditLogEntry.ACTION_CHANGED)
    except Exception as e:
        logger.exception(e)


def log_delete(sender, instance, **kwargs):
    """Log model delete"""
    try:
        if instance.pk is not None:
            changes = model_instance_diff(instance, None)
            if changes:
                create_log(instance, changes, AuditLogEntry.ACTION_DELETED)
    except Exception as e:
        logger.exception(e)


def init_auditlog():
    """Init auditlog"""
    for config in models_config.get_all_configs(False):
        if config.auditlog_disable or (not config.is_trionyx_model and not config.has_config('auditlog_disable')):
            continue

        post_save.connect(log_add, sender=config.model, dispatch_uid=(log_add, config.model, post_save))
        pre_save.connect(log_change, sender=config.model, dispatch_uid=(log_change, config.model, pre_save))
        post_delete.connect(log_delete, sender=config.model, dispatch_uid=(log_delete, config.model, post_delete))

        @tabs.register(config.model, code='history', name=_('History'), order=999)
        def layout(obj):
            return auditlog_layout(obj)



"""[[verbose_name]] app configuration"""
from trionyx.trionyx.apps import BaseConfig


class Config(BaseConfig):
    """[[verbose_name]] configuration"""

    name = 'apps.[[name]]'
    verbose_name = '[[verbose_name]]'



"""
trionyx.trionyx.url
~~~~~~~~~~~~~~~~~~~

:copyright: 2018 by Maikel Martens
:license: GPLv3
"""
from django.conf import settings
from django.conf.urls.static import static
from django.urls import path, include

from trionyx import views as core_views
from trionyx.trionyx import views
from trionyx.api.routers import AutoRouter

app_name = 'trionyx'

urlpatterns = [
    path('api/', include(AutoRouter().urls)),

    path('login/', views.LoginView.as_view(), name='login'),
    path('logout/', views.logout, name='logout'),

    path('account/edit/', views.UpdateUserAccountView.as_view(), name='edit-account'),
    path('account/view', views.ViewUserAccountView.as_view(), name='view-account'),

    # Global search
    path('global-search', views.GlobalSearchJsendView.as_view(), name='global-search'),

    # Tasks
    path('user-tasks', views.UserTasksJsend.as_view(), name='user-tasks'),

    # Dashboard
    path('', views.DashboardView.as_view(), name='dashboard'),
    path('dashboard/save/', views.SaveDashboardJsendView.as_view(), name='dashboard-save'),
    path('dashboard/widget-data/', views.WidgetDataJsendView.as_view(), name='dashboard-widget-data'),
    path('dashboard/widget-config/<str:code>/', views.WidgetConfigDialog.as_view(), name='dashboard-widget-config'),

    path('model-filter-fields/', views.FilterFieldsJsendView.as_view(), name='model-filter-fields'),

    # Generic model views
    path('model/<str:app>/<str:model>/', core_views.ListView.as_view(), name='model-list'),
    path('model/<str:app>/<str:model>/ajax/', core_views.ListJsendView.as_view(), name='model-list-ajax'),
    path('model/<str:app>/<str:model>/download/', core_views.ListExportView.as_view(), name='model-list-download'),
    path('model/<str:app>/<str:model>/choices/', core_views.ListChoicesJsendView.as_view(), name='model-list-choices'),

    path('model/<str:app>/<str:model>/create/', core_views.CreateView.as_view(), name='model-create'),
    path('model/<str:app>/<str:model>/create/<str:code>/', core_views.CreateView.as_view(), name='model-create-custom'),
    path('model/<str:app>/<str:model>/<int:pk>/', core_views.DetailTabView.as_view(), name='model-view'),
    path('model/<str:app>/<str:model>/<int:pk>/tab/', core_views.DetailTabJsendView.as_view(), name='model-tab'),
    path('model/<str:app>/<str:model>/<int:pk>/layout/<str:code>/', core_views.LayoutView.as_view(), name='model-view-custom'),

    path('model/<str:app>/<str:model>/<int:pk>/edit/', core_views.UpdateView.as_view(), name='model-edit'),
    path('model/<str:app>/<str:model>/<int:pk>/edit/<str:code>/', core_views.UpdateView.as_view(), name='model-edit-custom'),
    path('model/<str:app>/<str:model>/<int:pk>/delete/', core_views.DeleteView.as_view(), name='model-delete'),

    # Generic Dialog views
    path('dialog/model/<str:app>/<str:model>/<int:pk>/layout/<str:code>/', core_views.LayoutDialog.as_view(), name='model-dialog-view'),
    path('dialog/model/<str:app>/<str:model>/create/', core_views.CreateDialog.as_view(), name='model-dialog-create'),
    path('dialog/model/<str:app>/<str:model>/create/<str:code>/', core_views.CreateDialog.as_view(), name='model-dialog-create-custom'),
    path('dialog/model/<str:app>/<str:model>/<int:pk>/edit/', core_views.UpdateDialog.as_view(), name='model-dialog-edit'),
    path(
        'dialog/model/<str:app>/<str:model>/<int:pk>/edit/<str:code>/', core_views.UpdateDialog.as_view(),
        name='model-dialog-edit-custom'
    ),
    path('dialog/model/<str:app>/<str:model>/<int:pk>/delete/', core_views.DeleteDialog.as_view(), name='model-dialog-delete'),

    # Mass actions
    path('mass/<str:app>/<str:model>/delete', views.MassDeleteDialog.as_view(), name='model-mass-delete'),
    path('mass/<str:app>/<str:model>/update', views.MassUpdateView.as_view(), name='model-mass-update'),
]

if settings.DEBUG:
    urlpatterns += static(
        settings.MEDIA_URL, document_root=settings.MEDIA_ROOT
    )
else:
    from trionyx.trionyx.views import media_nginx_accel
    urlpatterns += [
        path('media/<path:path>', media_nginx_accel),
    ]



"""
trionyx.forms.layout
~~~~~~~~~~~~~~~~~~~~

:copyright: 2018 by Maikel Martens
:license: GPLv3
"""
from crispy_forms.layout import *  # noqa F403
from crispy_forms.bootstrap import *  # noqa F403
from django.template.loader import render_to_string
from django.contrib.contenttypes.models import ContentType
from trionyx import utils

from trionyx.utils import (
    get_current_locale,
    get_datetime_input_format,
    datetime_format_to_django_template,
    datetime_format_to_momentjs
)


class HtmlTemplate:
    """HTML template renderer"""

    def __init__(self, template, context=None):
        """Init layout object"""
        self.template = template
        self.context = context if context else {}

    def render(self, form, form_style, context, template_pack=TEMPLATE_PACK, **kwargs):
        """Render template"""
        return render_to_string(self.template, {
            **context.flatten(),
            **self.context,
        })


class DateTimePicker(Field):
    """DatetimePicker field renderer"""

    template = 'trionyx/forms/datetimepicker.html'

    glyphicon = 'glyphicon-calendar'

    locale = False
    """Locale of datetime picker default is active locale"""

    format = False
    """Date format, use python format: %Y-%m-%d %H:%M:%S"""

    day_view_header_format = 'MMMM YYYY'
    """Changes the heading of the datepicker when in "days" view."""

    stepping = 5
    """Number of minutes the up/down arrow's will move the minutes value in the time picker"""

    min_date = False
    """Prevents date/time selections before this date."""

    max_date = False
    """Prevents date/time selections after this date."""

    use_current = True
    """On show, will set the picker to the current date/time"""

    default_date = False
    """Sets the picker default date/time. Overrides useCurrent"""

    side_by_side = True
    """Shows the picker side by side when using the time and date together."""

    view_mode = 'days'
    """The default view to display when the picker is shown: 'decades','years','months','days'"""

    toolbar_placement = 'bottom'
    """Changes the placement of the icon toolbar: 'default', 'top', 'bottom'"""

    show_today_button = True
    """Show the "Today" button in the icon toolbar. Clicking the "Today" button will set the calendar view and set the date to now."""

    show_clear = True
    """Show the "Clear" button in the icon toolbar. Clicking the "Clear" button will set the calendar to null."""

    show_close = False
    """Show the "Close" button in the icon toolbar."""

    valid_options = [
        'locale',
        'format',
        'day_view_header_format',
        'stepping',
        'min_date',
        'max_date',
        'use_current',
        'default_date',
        'side_by_side',
        'view_mode',
        'toolbar_placement',
        'show_today_button',
        'show_clear',
        'show_close'
    ]

    def __init__(self, field, **kwargs):
        """Init DateTimePicker"""
        if not self.format:
            self.format = get_datetime_input_format()

        if not self.locale:
            self.locale = get_current_locale()

        for key in self.valid_options:
            value = kwargs.pop(key, getattr(self, key))
            if isinstance(value, bool):
                value = str(value).lower()
            kwargs['data-{}'.format(key)] = str(value if key != 'format' else datetime_format_to_momentjs(value))

        # Disable autocomplete default
        if 'autocomplete' not in kwargs:
            kwargs['autocomplete'] = 'off'

        kwargs['css_class'] = '{} datetimepicker datetimeinput form-control'.format(kwargs.get('css_class', '')).strip()
        super().__init__(field, **kwargs)

    def render(self, *args, **kwargs):
        """Render field"""
        extra_context = kwargs.get('extra_context', {})
        extra_context['glyphicon'] = self.glyphicon
        extra_context['input_format'] = datetime_format_to_django_template(self.format)
        kwargs['extra_context'] = extra_context
        return super().render(*args, **kwargs)


class TimePicker(DateTimePicker):
    """Timepicker field renderer"""

    format = '%H:%M'
    show_today_button = False
    glyphicon = 'glyphicon-time'


class InlineForm(LayoutObject):
    """Layout renderer for inline forms"""

    template = 'trionyx/forms/inlineform.html'

    def __init__(self, form_name, template=None):
        """Init Formset"""
        self.form_name = form_name
        if template:
            self.template = template

    def render(self, form, form_style, context, template_pack=TEMPLATE_PACK):
        """Render form"""
        return render_to_string(self.template, {
            'inline_form': form.get_inline_forms()[self.form_name]
        }, utils.get_current_request())


class Filters:
    """Form Filter field"""

    def __init__(self, name, model=None, content_type_input_id=None):
        """Init Filter field, model or content_type_input_id must be supplied"""
        self.name = name
        self.model = model
        self.content_type_input_id = content_type_input_id

        if not model and not content_type_input_id:
            raise Exception('A model or content_type_input_id must be supplied')

    def render(self, form, form_style, context, **kwargs):
        """Render template"""
        value = form.data.get(self.name, form.initial.get(self.name, '[]'))
        if not value:
            value = '[]'

        return render_to_string('trionyx/forms/filters.html', {
            **context.flatten(),
            'uuid': utils.random_string(8),
            'name': self.name,
            'value': value,
            'content_type_input_id': self.content_type_input_id,
            'content_type_id': ContentType.objects.get_for_model(self.model).id if self.model else -1
        })



"""Project Celery app"""
import os

from celery import Celery
from celery.signals import after_setup_logger
from trionyx.log import enable_db_logger

# set the default Django settings module for the 'celery' program.
os.environ.setdefault('DJANGO_SETTINGS_MODULE', 'config.settings')

# Create celery app
app = Celery('trionyx')

# Get celery config from django settings
app.config_from_object('django.conf:settings', namespace='CELERY')

# Load task modules from all registered Django app configs.
app.autodiscover_tasks()


@after_setup_logger.connect
def setup_loggers(logger, *args, **kwargs):
    """Enable DB logger for celery"""
    enable_db_logger()



from rest_framework.viewsets import ViewSet
from rest_framework.response import Response


class TestApi(ViewSet):


    def list(self, request, format=None):
        """
        Return a list of all users.
        """
        return Response('test')


"""
trionyx.trionyx.templatetags.trionyx
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

:copyright: 2018 by Maikel Martens
:license: GPLv3
"""
import json
from datetime import date

from django import template
from django.core.serializers import serialize
from django.db.models.query import QuerySet
from django.utils.safestring import mark_safe

register = template.Library()


@register.simple_tag
def active_menu_item(request, item):
    """Check if given menu item is active"""
    if item.is_active(request.path):
        return 'active'
    return ''


@register.simple_tag(takes_context=True)
def render_component(context, component):
    """Render given component"""
    return component.render(context.flatten(), context.get('request'))


@register.filter
def jsonify(obj):
    """Jsonify given input"""
    if isinstance(obj, QuerySet):
        return mark_safe(serialize(obj))
    return mark_safe(json.dumps(obj))


@register.simple_tag
def model_url(model, view_name, code=None):
    """Short cut for generating model urls"""
    from trionyx.urls import model_url
    return model_url(model, view_name, code)


@register.filter
def is_date(value):
    """Check if value is date object"""
    return isinstance(value, date)



from setuptools import setup
from setuptools import find_packages
from trionyx import __version__

def readme():
    with open('README.rst') as f:
        return f.read()

setup(
    name = 'Trionyx',
    packages = find_packages(exclude=['app']),
    include_package_data=True,
    scripts=['trionyx/bin/trionyx'],
    version = __version__,
    description = 'Trionyx is an application framework for managing data and processes',
    long_description=readme(),
    classifiers=[
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python :: 3',
        'Framework :: Django',
    ],
    author = 'Maikel Martens',
    author_email = 'maikel@martens.me',
    license='GPL3',
    url = 'https://github.com/krukas/Trionyx',
    download_url = 'https://github.com/krukas/Trionyx/releases/tag/{}'.format(__version__),
    keywords = ['Django', 'Trionyx', 'framework', 'admin', 'data', 'processes', 'application', 'stack'],
    python_requires='~=3.5',
    install_requires=[
        "Django >= 2.2.0, < 2.3",

        # Django apps
        'django-crispy-forms >= 1.7.2, < 1.8',
        'django_compressor >= 2.3.0, < 2.4',
        'django_jsend==0.4',
        'jsonfield2==3.0.2',
        'django-watson >= 1.5.0, < 1.6',
        'djangorestframework >= 3.10.0, < 3.11',

        'Babel >= 2.6.0, < 2.7',

        'celery >= 4.3.0, < 4.4',

        'Pillow>=5.0.0',
    ],
    extras_require={
        'dev': [
            'colorlog',
            'django-extensions',
            'django-debug-toolbar',
            'Werkzeug',
            'coverage',
            'flake8',
            'pydocstyle',
            'ipython',
            'Sphinx',
            'sphinx_rtd_theme',
        ]
    }
)
